from __future__ import annotations

import pytest

from pybind11_tests import ConstructorStats, UserType
from pybind11_tests import stl as m


def test_vector(doc):
    """std::vector <-> list"""
    lst = m.cast_vector()
    assert lst == [1]
    lst.append(2)
    assert m.load_vector(lst)
    assert m.load_vector(tuple(lst))

    assert m.cast_bool_vector() == [True, False]
    assert m.load_bool_vector([True, False])
    assert m.load_bool_vector((True, False))

    assert doc(m.cast_vector) == "cast_vector() -> list[int]"
    assert doc(m.load_vector) == "load_vector(arg0: list[int]) -> bool"

    # Test regression caused by 936: pointers to stl containers weren't castable
    assert m.cast_ptr_vector() == ["lvalue", "lvalue"]


def test_deque():
    """std::deque <-> list"""
    lst = m.cast_deque()
    assert lst == [1]
    lst.append(2)
    assert m.load_deque(lst)
    assert m.load_deque(tuple(lst))


def test_array(doc):
    """std::array <-> list"""
    lst = m.cast_array()
    assert lst == [1, 2]
    assert m.load_array(lst)
    assert m.load_array(tuple(lst))

    assert doc(m.cast_array) == "cast_array() -> Annotated[list[int], FixedSize(2)]"
    assert (
        doc(m.load_array)
        == "load_array(arg0: Annotated[list[int], FixedSize(2)]) -> bool"
    )


def test_valarray(doc):
    """std::valarray <-> list"""
    lst = m.cast_valarray()
    assert lst == [1, 4, 9]
    assert m.load_valarray(lst)
    assert m.load_valarray(tuple(lst))

    assert doc(m.cast_valarray) == "cast_valarray() -> list[int]"
    assert doc(m.load_valarray) == "load_valarray(arg0: list[int]) -> bool"


def test_map(doc):
    """std::map <-> dict"""
    d = m.cast_map()
    assert d == {"key": "value"}
    assert "key" in d
    d["key2"] = "value2"
    assert "key2" in d
    assert m.load_map(d)

    assert doc(m.cast_map) == "cast_map() -> dict[str, str]"
    assert doc(m.load_map) == "load_map(arg0: dict[str, str]) -> bool"


def test_set(doc):
    """std::set <-> set"""
    s = m.cast_set()
    assert s == {"key1", "key2"}
    s.add("key3")
    assert m.load_set(s)
    assert m.load_set(frozenset(s))

    assert doc(m.cast_set) == "cast_set() -> set[str]"
    assert doc(m.load_set) == "load_set(arg0: set[str]) -> bool"


def test_recursive_casting():
    """Tests that stl casters preserve lvalue/rvalue context for container values"""
    assert m.cast_rv_vector() == ["rvalue", "rvalue"]
    assert m.cast_lv_vector() == ["lvalue", "lvalue"]
    assert m.cast_rv_array() == ["rvalue", "rvalue", "rvalue"]
    assert m.cast_lv_array() == ["lvalue", "lvalue"]
    assert m.cast_rv_map() == {"a": "rvalue"}
    assert m.cast_lv_map() == {"a": "lvalue", "b": "lvalue"}
    assert m.cast_rv_nested() == [[[{"b": "rvalue", "c": "rvalue"}], [{"a": "rvalue"}]]]
    assert m.cast_lv_nested() == {
        "a": [[["lvalue", "lvalue"]], [["lvalue", "lvalue"]]],
        "b": [[["lvalue", "lvalue"], ["lvalue", "lvalue"]]],
    }

    # Issue #853 test case:
    z = m.cast_unique_ptr_vector()
    assert z[0].value == 7
    assert z[1].value == 42


def test_move_out_container():
    """Properties use the `reference_internal` policy by default. If the underlying function
    returns an rvalue, the policy is automatically changed to `move` to avoid referencing
    a temporary. In case the return value is a container of user-defined types, the policy
    also needs to be applied to the elements, not just the container."""
    c = m.MoveOutContainer()
    moved_out_list = c.move_list
    assert [x.value for x in moved_out_list] == [0, 1, 2]


@pytest.mark.skipif(not hasattr(m, "has_optional"), reason="no <optional>")
def test_optional():
    assert m.double_or_zero(None) == 0
    assert m.double_or_zero(42) == 84
    pytest.raises(TypeError, m.double_or_zero, "foo")

    assert m.half_or_none(0) is None
    assert m.half_or_none(42) == 21
    pytest.raises(TypeError, m.half_or_none, "foo")

    assert m.test_nullopt() == 42
    assert m.test_nullopt(None) == 42
    assert m.test_nullopt(42) == 42
    assert m.test_nullopt(43) == 43

    assert m.test_no_assign() == 42
    assert m.test_no_assign(None) == 42
    assert m.test_no_assign(m.NoAssign(43)) == 43
    pytest.raises(TypeError, m.test_no_assign, 43)

    assert m.nodefer_none_optional(None)

    holder = m.OptionalHolder()
    mvalue = holder.member
    assert mvalue.initialized
    assert holder.member_initialized()

    props = m.OptionalProperties()
    assert int(props.access_by_ref) == 42
    assert int(props.access_by_copy) == 42


@pytest.mark.skipif(
    not hasattr(m, "has_exp_optional"), reason="no <experimental/optional>"
)
def test_exp_optional():
    assert m.double_or_zero_exp(None) == 0
    assert m.double_or_zero_exp(42) == 84
    pytest.raises(TypeError, m.double_or_zero_exp, "foo")

    assert m.half_or_none_exp(0) is None
    assert m.half_or_none_exp(42) == 21
    pytest.raises(TypeError, m.half_or_none_exp, "foo")

    assert m.test_nullopt_exp() == 42
    assert m.test_nullopt_exp(None) == 42
    assert m.test_nullopt_exp(42) == 42
    assert m.test_nullopt_exp(43) == 43

    assert m.test_no_assign_exp() == 42
    assert m.test_no_assign_exp(None) == 42
    assert m.test_no_assign_exp(m.NoAssign(43)) == 43
    pytest.raises(TypeError, m.test_no_assign_exp, 43)

    holder = m.OptionalExpHolder()
    mvalue = holder.member
    assert mvalue.initialized
    assert holder.member_initialized()

    props = m.OptionalExpProperties()
    assert int(props.access_by_ref) == 42
    assert int(props.access_by_copy) == 42


@pytest.mark.skipif(not hasattr(m, "has_boost_optional"), reason="no <boost/optional>")
def test_boost_optional():
    assert m.double_or_zero_boost(None) == 0
    assert m.double_or_zero_boost(42) == 84
    pytest.raises(TypeError, m.double_or_zero_boost, "foo")

    assert m.half_or_none_boost(0) is None
    assert m.half_or_none_boost(42) == 21
    pytest.raises(TypeError, m.half_or_none_boost, "foo")

    assert m.test_nullopt_boost() == 42
    assert m.test_nullopt_boost(None) == 42
    assert m.test_nullopt_boost(42) == 42
    assert m.test_nullopt_boost(43) == 43

    assert m.test_no_assign_boost() == 42
    assert m.test_no_assign_boost(None) == 42
    assert m.test_no_assign_boost(m.NoAssign(43)) == 43
    pytest.raises(TypeError, m.test_no_assign_boost, 43)

    holder = m.OptionalBoostHolder()
    mvalue = holder.member
    assert mvalue.initialized
    assert holder.member_initialized()

    props = m.OptionalBoostProperties()
    assert int(props.access_by_ref) == 42
    assert int(props.access_by_copy) == 42


def test_reference_sensitive_optional():
    assert m.double_or_zero_refsensitive(None) == 0
    assert m.double_or_zero_refsensitive(42) == 84
    pytest.raises(TypeError, m.double_or_zero_refsensitive, "foo")

    assert m.half_or_none_refsensitive(0) is None
    assert m.half_or_none_refsensitive(42) == 21
    pytest.raises(TypeError, m.half_or_none_refsensitive, "foo")

    assert m.test_nullopt_refsensitive() == 42
    assert m.test_nullopt_refsensitive(None) == 42
    assert m.test_nullopt_refsensitive(42) == 42
    assert m.test_nullopt_refsensitive(43) == 43

    assert m.test_no_assign_refsensitive() == 42
    assert m.test_no_assign_refsensitive(None) == 42
    assert m.test_no_assign_refsensitive(m.NoAssign(43)) == 43
    pytest.raises(TypeError, m.test_no_assign_refsensitive, 43)

    holder = m.OptionalRefSensitiveHolder()
    mvalue = holder.member
    assert mvalue.initialized
    assert holder.member_initialized()

    props = m.OptionalRefSensitiveProperties()
    assert int(props.access_by_ref) == 42
    assert int(props.access_by_copy) == 42


@pytest.mark.skipif(not hasattr(m, "has_filesystem"), reason="no <filesystem>")
def test_fs_path():
    from pathlib import Path

    class PseudoStrPath:
        def __fspath__(self):
            return "foo/bar"

    class PseudoBytesPath:
        def __fspath__(self):
            return b"foo/bar"

    assert m.parent_path(Path("foo/bar")) == Path("foo")
    assert m.parent_path("foo/bar") == Path("foo")
    assert m.parent_path(b"foo/bar") == Path("foo")
    assert m.parent_path(PseudoStrPath()) == Path("foo")
    assert m.parent_path(PseudoBytesPath()) == Path("foo")


@pytest.mark.skipif(not hasattr(m, "load_variant"), reason="no <variant>")
def test_variant(doc):
    assert m.load_variant(1) == "int"
    assert m.load_variant("1") == "std::string"
    assert m.load_variant(1.0) == "double"
    assert m.load_variant(None) == "std::nullptr_t"

    assert m.load_variant_2pass(1) == "int"
    assert m.load_variant_2pass(1.0) == "double"

    assert m.cast_variant() == (5, "Hello")

    assert (
        doc(m.load_variant) == "load_variant(arg0: Union[int, str, float, None]) -> str"
    )


@pytest.mark.skipif(
    not hasattr(m, "load_monostate_variant"), reason="no std::monostate"
)
def test_variant_monostate(doc):
    assert m.load_monostate_variant(None) == "std::monostate"
    assert m.load_monostate_variant(1) == "int"
    assert m.load_monostate_variant("1") == "std::string"

    assert m.cast_monostate_variant() == (None, 5, "Hello")

    assert (
        doc(m.load_monostate_variant)
        == "load_monostate_variant(arg0: Union[None, int, str]) -> str"
    )


def test_vec_of_reference_wrapper():
    """#171: Can't return reference wrappers (or STL structures containing them)"""
    assert (
        str(m.return_vec_of_reference_wrapper(UserType(4)))
        == "[UserType(1), UserType(2), UserType(3), UserType(4)]"
    )


def test_stl_pass_by_pointer(msg):
    """Passing nullptr or None to an STL container pointer is not expected to work"""
    with pytest.raises(TypeError) as excinfo:
        m.stl_pass_by_pointer()  # default value is `nullptr`
    assert (
        msg(excinfo.value)
        == """
        stl_pass_by_pointer(): incompatible function arguments. The following argument types are supported:
            1. (v: list[int] = None) -> list[int]

        Invoked with:
    """
    )

    with pytest.raises(TypeError) as excinfo:
        m.stl_pass_by_pointer(None)
    assert (
        msg(excinfo.value)
        == """
        stl_pass_by_pointer(): incompatible function arguments. The following argument types are supported:
            1. (v: list[int] = None) -> list[int]

        Invoked with: None
    """
    )

    assert m.stl_pass_by_pointer([1, 2, 3]) == [1, 2, 3]


def test_missing_header_message():
    """Trying convert `list` to a `std::vector`, or vice versa, without including
    <pybind11/stl.h> should result in a helpful suggestion in the error message"""
    import pybind11_cross_module_tests as cm

    expected_message = (
        "Did you forget to `#include <pybind11/stl.h>`? Or <pybind11/complex.h>,\n"
        "<pybind11/functional.h>, <pybind11/chrono.h>, etc. Some automatic\n"
        "conversions are optional and require extra headers to be included\n"
        "when compiling your pybind11 module."
    )

    with pytest.raises(TypeError) as excinfo:
        cm.missing_header_arg([1.0, 2.0, 3.0])
    assert expected_message in str(excinfo.value)

    with pytest.raises(TypeError) as excinfo:
        cm.missing_header_return()
    assert expected_message in str(excinfo.value)


def test_function_with_string_and_vector_string_arg():
    """Check if a string is NOT implicitly converted to a list, which was the
    behavior before fix of issue #1258"""
    assert m.func_with_string_or_vector_string_arg_overload(("A", "B")) == 2
    assert m.func_with_string_or_vector_string_arg_overload(["A", "B"]) == 2
    assert m.func_with_string_or_vector_string_arg_overload("A") == 3


def test_stl_ownership():
    cstats = ConstructorStats.get(m.Placeholder)
    assert cstats.alive() == 0
    r = m.test_stl_ownership()
    assert len(r) == 1
    del r
    assert cstats.alive() == 0


def test_array_cast_sequence():
    assert m.array_cast_sequence((1, 2, 3)) == [1, 2, 3]


def test_issue_1561():
    """check fix for issue #1561"""
    bar = m.Issue1561Outer()
    bar.list = [m.Issue1561Inner("bar")]
    assert bar.list
    assert bar.list[0].data == "bar"


def test_return_vector_bool_raw_ptr():
    # Add `while True:` for manual leak checking.
    v = m.return_vector_bool_raw_ptr()
    assert isinstance(v, list)
    assert len(v) == 4513



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

_selftype = None


class selftype:  # noqa: N801
    """When used in EUDFuncMethod's type declaration, This is interpreted
    as the owning class itself
    """

    @staticmethod
    def cast(_from):
        return _selftype.cast(_from)


def _set_selftype(t):
    global _selftype
    _selftype = t



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from .constexpr import ConstExpr, Evaluate, Forward, IsConstExpr
from .payload import (
    CompressPayload,
    CreatePayload,
    GetObjectAddr,
    RegisterCreatePayloadCallback,
    ShufflePayload,
)
from .rlocint import RlocInt, RlocInt_C, toRlocInt

__all__ = [
    "ConstExpr",
    "Evaluate",
    "Forward",
    "IsConstExpr",
    "CompressPayload",
    "CreatePayload",
    "GetObjectAddr",
    "RegisterCreatePayloadCallback",
    "ShufflePayload",
    "RlocInt",
    "RlocInt_C",
    "toRlocInt",
]



from random import sample, shuffle

from helper import *


@TestInstance
def main():
    i = EUDVariable()
    s = EUDVariable()

    if EUDWhile()(i <= 11 - 1):
        i += 1

        EUDSwitch(i)
        if EUDSwitchCase()(7):
            s += 1

        if EUDSwitchCase()(3):
            s += 10

        if EUDSwitchCase()(4, 6):
            s += 100
            EUDBreak()

        if EUDSwitchCase()(1, 2, 5, 8):
            s += 1000
            EUDBreak()

        # negative number is allowed only with full-ranged mask
        if EUDSwitchCase()(-1, -3, -7):
            s += -666
            EUDBreak()

        if EUDSwitchDefault()():
            s += 10000

        EUDEndSwitch()
    EUDEndWhile()
    test_equality("Switch test", s, 34421)

    random_numbers = sample(range(8192), 32)
    random_cases = random_numbers[:16]

    @EUDFunc
    def branch(x):
        ret = EUDVariable()
        ret << 0
        EUDSwitch(x, 8191)
        for case in random_cases:
            if EUDSwitchCase()(case):
                ret << 3 * case
                EUDBreak()
        if EUDSwitchDefault()():
            ret << -1
        EUDEndSwitch()
        return ret

    shuffle(random_numbers)
    ra = EUDArray(random_numbers)
    SetVariables([s, i], [0, 0])
    if EUDWhileNot()(i >= 32):
        x = branch(ra[i])
        SetVariables([i, s], [1, x], [Add, Add])
    EUDEndWhile()

    result = (-16 + 3 * sum(random_cases)) & 0xFFFFFFFF

    test_equality("Switch binary search test", s, result)



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from __future__ import annotations

import time
from typing import TYPE_CHECKING, TypeAlias

from ...bindings._rust import allocator
from ...localize import _
from ...utils import EPError, _rand_lst, ep_assert
from .constexpr import Evaluable, Evaluate, Forward
from .pbuffer import Payload
from .rlocint import RlocInt, RlocInt_C

if TYPE_CHECKING:
    from ..eudobj import EUDObject

_found_objects_dict: dict[EUDObject, int] = {}
_untraversed_objects: list[EUDObject] = []
_dynamic_objects_set: set[EUDObject] = set()
_payload_builder = allocator.PayloadBuilder()

PHASE_COLLECTING = 1
PHASE_ALLOCATING = 2
PHASE_WRITING = 3
phase: int = 0

_payload_compress: bool = False
_payload_shuffle: bool = True

# -------

_last_time: float = 0.0
_is_logging_enabled: bool = False


def _set_payload_logger_mode(mode: bool) -> None:
    global _is_logging_enabled
    _is_logging_enabled = mode


def lprint(text: str, flush: bool = False):
    global _last_time, _is_logging_enabled
    if not _is_logging_enabled:
        return

    current_time = time.time()
    if flush or (current_time - _last_time) >= 0.5:
        _last_time = current_time
        print(text, end="\n" if flush else "\r")


def CompressPayload(mode: bool) -> None:  # noqa: N802
    """Set payload compression mode.

    :param mode: If true, enable object stacking (compression). If false,
    disable it.
    """
    global _payload_compress
    ep_assert(mode in (True, False), _("Invalid type") + f": {mode}")
    _payload_compress = True if mode else False


def ShufflePayload(mode: bool) -> None:  # noqa: N802
    """Set payload shuffle mode.

    :param mode: If true, enable object shuffling (compression). If false,
    disable it.
    """
    global _payload_shuffle
    ep_assert(mode in (True, False), _("Invalid type") + f": {mode}")
    _payload_shuffle = True if mode else False


class ObjCollector:
    """
    Object having PayloadBuffer-like interfaces. Collects all objects by
    calling object.CollectDependency() for every related objects.
    """

    def __init__(self) -> None:
        pass

    def StartWrite(self) -> None:  # noqa: N802
        pass

    def EndWrite(self) -> None:  # noqa: N802
        pass

    def WriteByte(self, number: int) -> None:  # noqa: N802
        pass

    def WriteWord(self, number: int) -> None:  # noqa: N802
        pass

    def WriteDword(self, obj: Evaluable) -> None:  # noqa: N802
        if not isinstance(obj, int):
            Evaluate(obj)

    def WritePack(self, structformat: str, arglist: list[Evaluable]) -> None:  # noqa: N802
        for arg in arglist:
            if not isinstance(arg, int):
                Evaluate(arg)

    def WriteBytes(self, b: bytes) -> None:  # noqa: N802
        pass

    def WriteSpace(self, spacesize: int) -> None:  # noqa: N802
        pass


def _collect_objects(root: EUDObject | Forward) -> None:
    global phase
    global _found_objects_dict
    global _dynamic_objects_set
    global _untraversed_objects

    lprint(_("[Stage 1/3] CollectObjects"), flush=True)

    phase = PHASE_COLLECTING

    objc = ObjCollector()
    _found_objects_dict = {}
    _dynamic_objects_set = set()
    _untraversed_objects = []

    # Evaluate root to register root object.
    # root may not have WritePayload() method e.g: Forward()
    Evaluate(root)

    while _untraversed_objects:
        while _untraversed_objects:
            lprint(
                _(" - Collected {} / {} objects").format(
                    len(_found_objects_dict),
                    len(_found_objects_dict) + len(_untraversed_objects),
                )
            )

            obj = _untraversed_objects.pop()

            objc.StartWrite()
            obj.CollectDependency(objc)
            objc.EndWrite()

        # Check for new objects
        for obj in _dynamic_objects_set:
            objc.StartWrite()
            obj.CollectDependency(objc)
            objc.EndWrite()

    if len(_found_objects_dict) == 0:
        raise EPError(_("No object collected"))

    if _payload_shuffle:
        # Shuffle objects -> Randomize(?) addresses
        iterobj = iter(_found_objects_dict)
        rootobj = next(iterobj)
        found_objects = _rand_lst(iterobj)
        found_objects.append(rootobj)
        _found_objects_dict = {
            obj: i for i, obj in enumerate(reversed(found_objects))
        }

    # cleanup
    phase = 0

    # Final
    lprint(
        _(" - Collected {} / {} objects").format(
            len(_found_objects_dict), len(_found_objects_dict)
        ),
        flush=True,
    )
    if not _payload_shuffle:
        lprint(_("ShufflePayload has been turned off."), flush=True)


# -------


def _allocate_objects() -> None:
    global phase
    global _payload_builder

    phase = PHASE_ALLOCATING
    lprint(_("[Stage 2/3] AllocObjects"), flush=True)

    if not _payload_compress:
        raise EPError("CompressPayload(False) is currently not supported")

    _payload_builder.alloc_objects(_found_objects_dict)

    phase = 0


def _construct_payload() -> Payload:
    global phase
    global _payload_builder

    phase = PHASE_WRITING
    lprint(_("[Stage 3/3] ConstructPayload"), flush=True)
    if _payload_builder is None:
        raise EPError(_("PayloadBuilder is not instantiated"))

    payload = _payload_builder.construct_payload(_found_objects_dict)
    phase = 0
    return Payload(*payload)


RegisterCreatePayloadCallback = _payload_builder.register_create_payload_callback
_register_after_collecting_callback = (
    _payload_builder.register_after_collecting_callback
)


def CreatePayload(root: EUDObject | Forward) -> Payload:  # noqa: N802
    from ..rawtrigger.rawtriggerdef import RawTrigger

    # Call callbacks
    _payload_builder.call_callbacks_on_create_payload()
    _collect_objects(root)
    _payload_builder.call_callbacks_after_collecting()
    setattr(RawTrigger, "WritePayload", RawTrigger._allocate_trigger)
    _allocate_objects()
    setattr(RawTrigger, "WritePayload", RawTrigger._write_trigger)
    return _construct_payload()


_PayloadBuffer: TypeAlias = allocator.ObjAllocator | allocator.PayloadBuffer
defri: RlocInt_C = RlocInt(0, 4)


def GetObjectAddr(obj: EUDObject) -> RlocInt_C:  # noqa: N802
    global _payload_builder
    global _found_objects_dict
    global _untraversed_objects
    global _dynamic_objects_set

    if phase == PHASE_COLLECTING:
        if obj not in _found_objects_dict:
            _untraversed_objects.append(obj)
            _found_objects_dict[obj] = len(_found_objects_dict)
            if obj.DynamicConstructed():
                _dynamic_objects_set.add(obj)

        return defri

    elif phase == PHASE_ALLOCATING:
        return defri

    elif phase == PHASE_WRITING:
        # ep_assert(_payload_builder.offset(_found_objects_dict[obj]) & 3 == 0)
        return _payload_builder.get_object_addr(_found_objects_dict[obj])

    else:
        raise EPError(_("Can't calculate object address now"))



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

import weakref
from dataclasses import dataclass

from typing_extensions import Self

from .. import core as c
from .. import ctrlstru as cs
from .. import utils as ut
from ..core.variable import eudv as ev
from ..localize import _
from .eudarray import EUDArray


class _ObjPoolData(c.ConstExpr):
    def __new__(cls, *args, **kwargs) -> Self:
        return super().__new__(cls, None)

    def __init__(self, size: int, max_fieldn: int = 8) -> None:
        super().__init__()
        self._vdict: weakref.WeakKeyDictionary = weakref.WeakKeyDictionary()
        self.size = size
        self.max_fieldn = max_fieldn

    def Evaluate(self):  # noqa: N802
        evb = ev.get_current_varbuffer()
        try:
            return evb._vdict[self].Evaluate()
        except KeyError:
            ret = evb.create_vartriggers(self, [0] * (self.max_fieldn * self.size))
            return ret.Evaluate()


class ObjPool:
    def __init__(self, size: int, max_fieldn: int = 8) -> None:
        self.size = size
        self.max_fieldn = max_fieldn
        self.remaining = c.EUDVariable(size)

        self.baseobj = _ObjPoolData(size, max_fieldn)
        self.data = EUDArray([72 * max_fieldn * i for i in range(size)])

    def full(self) -> c.Condition:
        return self.remaining == 0

    @c.EUDMethod
    def _alloc(self):
        """Allocate one object from pool"""
        if cs.EUDIf()(self.full()):
            c.EUDReturn(0)
        cs.EUDEndIf()

        self.remaining -= 1
        data = self.data[self.remaining]
        return data + self.baseobj

    def alloc(self, basetype, *args, **kwargs):
        ut.ep_assert(
            len(basetype._fielddict) <= self.max_fieldn,
            _("Only structs less than {} fields can be allocated").format(
                self.max_fieldn
            ),
        )
        data = self._alloc()
        data = basetype.cast(data)
        data.constructor(*args, **kwargs)
        return data

    def free(self, basetype, data) -> None:
        ut.ep_assert(
            len(basetype._fielddict) <= self.max_fieldn,
            _("Only structs less than {} fields can be allocated").format(
                self.max_fieldn
            ),
        )

        data = basetype.cast(data)
        data.destructor()
        self.data[self.remaining] = data - self.baseobj
        self.remaining += 1


@dataclass
class _GlobalObjPool:
    pool: ObjPool | None
    max_fieldn: int = 8
    max_object_num: int = 32768


_global_pool = _GlobalObjPool(pool=None, max_fieldn=8, max_object_num=32768)


def SetGlobalPoolFieldN(fieldn: int) -> None:  # noqa: N802
    global _global_pool
    ut.ep_assert(
        _global_pool.pool is None, "Global object pool is already initialized."
    )
    _global_pool.max_fieldn = fieldn


def get_global_pool() -> ObjPool:
    global _global_pool
    if _global_pool.pool is None:
        size, max_fieldn = _global_pool.max_object_num, _global_pool.max_fieldn
        _global_pool.pool = ObjPool(size, max_fieldn)
    return _global_pool.pool



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from ...localize import _
from ...utils import EPError
from .chktok import CHK
from .playerinfo import init_player_info
from .proptable import apply_property_map, init_property_map
from .stringmap import apply_string_map, init_stringmaps

_inited: bool = False
_chkt: CHK | None = None
_origchkt: CHK | None = None
_rawfile: bytes | None = None


def init_map_data(chkt: CHK, rawfile: bytes) -> None:
    global _inited, _origchkt, _chkt, _rawfile
    _chkt = chkt
    _origchkt = chkt.clone()
    _rawfile = rawfile

    init_stringmaps(chkt)
    init_property_map(chkt)
    init_player_info(chkt)
    _inited = True


def update_map_data() -> None:
    if _chkt is None:
        raise EPError(_("Must use LoadMap first"))
    apply_string_map(_chkt)
    apply_property_map(_chkt)


def IsMapdataInitialized() -> bool:  # noqa: N802
    return _inited


def GetChkTokenized() -> CHK:  # noqa: N802
    if _chkt is None:
        raise EPError(_("Must use LoadMap first"))
    return _chkt


def GetOriginalChkTokenized() -> CHK:  # noqa: N802
    """NEVER MODIFY ANYTHING WITHIN THIS CHKTOK"""
    if _origchkt is None:
        raise EPError(_("Must use LoadMap first"))
    return _origchkt


def GetRawFile() -> bytes:  # noqa: N802
    if _rawfile is None:
        raise EPError(_("Must use LoadMap first"))
    return _rawfile



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from ... import core as c
from ... import ctrlstru as cs
from ... import utils as ut

# Formula source : http://nghiaho.com/?p=997


@c.EUDFunc
def f_atan2(y, x):
    signflags = c.EUDVariable()
    signflags << 0

    # Check x sign
    c.RawTrigger(
        conditions=x >= 0x80000000,
        actions=[
            signflags.AddNumber(1),  # set xsign
            x.AddNumberX(0xFFFFFFFF, 0x55555555),
            x.AddNumberX(0xFFFFFFFF, 0xAAAAAAAA),
            x.AddNumber(1),
        ],
    )

    # Check y sign
    c.RawTrigger(
        conditions=y >= 0x80000000,
        actions=[
            signflags.AddNumber(2),  # set ysign
            y.AddNumberX(0xFFFFFFFF, 0x55555555),
            y.AddNumberX(0xFFFFFFFF, 0xAAAAAAAA),
            y.AddNumber(1),
        ],
    )

    # Check x/y order
    if cs.EUDIf()(y >= x):
        z = c.EUDXVariable(0, c.SetTo, 0)
        # Swap x, y so that y <= x
        c.SeqCompute(
            [
                (signflags, c.Add, 4),  # set xyabscmp
                (z, c.SetTo, x),
                (x, c.SetTo, y),
                (y, None, z),
            ]
        )
    cs.EUDEndIf()

    # To prevent overflow, we limit values of y and x.
    # atan value is maximized when x = y, then atan_value = 45 * x**3
    # 45 * x**3 <= 0xFFFFFFFF : x <= 456.99....
    if cs.EUDIf()(x >= 400):
        # Normalize below 400
        divn = x // 400
        divn += 1
        x //= divn
        y //= divn
    cs.EUDEndIf()

    # Calculate arctan value
    # arctan(z) ~= z * (45 - (z-1) * (14 + 4*z)), 0 <= z <= 1
    # arctan(y/x) ~= y/x * (45 - (y-x)/x * (14x + 4y)/x))
    # arctan(y/x) ~= y * (45*x*x - (y-x)(14x+4y)) / (x*x*x)
    t1 = x * x
    t2 = y * (45 * t1 - (y - x) * (14 * x + 4 * y))
    t3 = x * t1
    atan_value = t2 // t3

    # Translate angles by sign flags
    #
    #      |  0 |  1 | xsign          |  0 |  1 | xsign
    # -----+----+----+-----      -----+----+----+-----
    #   0  |  0+|180-|             0  | 90-| 90+|
    # -----+----+----+           -----+----+----+
    #   1  |360-|180+|             1  |270+|270-|
    # -----+----+----+           -----+----+----+
    # ysign|      xyabscmp=0     ysign|      xyabscmp=1

    cs.EUDSwitch(signflags, 7)
    swblock = ut.EUDGetLastBlock()[1]
    if cs.EUDSwitchCase()(0):  # xsign, ysign, xyabscmp = 0, 0, 0
        # atan_value
        c.SetNextTrigger(swblock["swend"])
    if cs.EUDSwitchCase()(1):  # xsign, ysign, xyabscmp = 1, 0, 0
        # 180 - atan_value
        c.RawTrigger(
            nextptr=swblock["swend"],
            actions=[
                atan_value.AddNumberX(0xFFFFFFFF, 0x55555555),
                atan_value.AddNumberX(0xFFFFFFFF, 0xAAAAAAAA),
                atan_value.AddNumber(180 + 1),
            ],
        )
    if cs.EUDSwitchCase()(2):  # xsign, ysign, xyabscmp = 0, 1, 0
        # 360 - atan_value
        c.RawTrigger(
            nextptr=swblock["swend"],
            actions=[
                atan_value.AddNumberX(0xFFFFFFFF, 0x55555555),
                atan_value.AddNumberX(0xFFFFFFFF, 0xAAAAAAAA),
                atan_value.AddNumber(360 + 1),
            ],
        )
    if cs.EUDSwitchCase()(3):  # xsign, ysign, xyabscmp = 1, 1, 0
        # 180 + atan_value
        c.RawTrigger(
            nextptr=swblock["swend"],
            actions=atan_value.AddNumber(180),
        )
    if cs.EUDSwitchCase()(4):  # xsign, ysign, xyabscmp = 0, 0, 1
        # 90 - atan_value
        c.RawTrigger(
            nextptr=swblock["swend"],
            actions=[
                atan_value.AddNumberX(0xFFFFFFFF, 0x55555555),
                atan_value.AddNumberX(0xFFFFFFFF, 0xAAAAAAAA),
                atan_value.AddNumber(90 + 1),
            ],
        )
    if cs.EUDSwitchCase()(5):  # xsign, ysign, xyabscmp = 1, 0, 1
        # 90 + atan_value
        c.RawTrigger(
            nextptr=swblock["swend"],
            actions=atan_value.AddNumber(90),
        )
    if cs.EUDSwitchCase()(6):  # xsign, ysign, xyabscmp = 0, 1, 1
        # 270 + atan_value
        c.RawTrigger(
            nextptr=swblock["swend"],
            actions=atan_value.AddNumber(270),
        )
    if cs.EUDSwitchCase()(7):  # xsign, ysign, xyabscmp = 1, 1, 1
        # 270 - atan_value
        c.RawTrigger(
            nextptr=swblock["swend"],
            actions=[
                atan_value.AddNumberX(0xFFFFFFFF, 0x55555555),
                atan_value.AddNumberX(0xFFFFFFFF, 0xAAAAAAAA),
                atan_value.AddNumber(270 + 1),
            ],
        )
    cs.EUDEndSwitch()
    return atan_value


@c.EUDFunc
def f_atan2_256(y, x):
    signflags = c.EUDVariable()
    signflags << 0

    # Check x sign
    c.RawTrigger(
        conditions=x >= 0x80000000,
        actions=[
            signflags.AddNumber(1),  # set xsign
            x.AddNumberX(0xFFFFFFFF, 0x55555555),
            x.AddNumberX(0xFFFFFFFF, 0xAAAAAAAA),
            x.AddNumber(1),
        ],
    )

    # Check y sign
    c.RawTrigger(
        conditions=y >= 0x80000000,
        actions=[
            signflags.AddNumber(2),  # set ysign
            y.AddNumberX(0xFFFFFFFF, 0x55555555),
            y.AddNumberX(0xFFFFFFFF, 0xAAAAAAAA),
            y.AddNumber(1),
        ],
    )

    # Check x/y order
    if cs.EUDIf()(y >= x):
        z = c.EUDXVariable(0, c.SetTo, 0)
        # Swap x, y so that y <= x
        c.SeqCompute(
            [
                (signflags, c.Add, 4),  # set xyabscmp
                (z, c.SetTo, x),
                (x, c.SetTo, y),
                (y, None, z),
            ]
        )
    cs.EUDEndIf()

    # To prevent overflow, we limit values of y and x.
    # atan value is maximized when x = y, then atan_value = 32 * 45 * x**3
    # 1440 * x**3 <= 0xFFFFFFFF : x <= 143.946....
    if cs.EUDIf()(x >= 128):
        # Normalize below 128
        divn = x // 128
        divn += 1
        x //= divn
        y //= divn
    cs.EUDEndIf()

    # Calculate arctan value
    # y * 32(45x² - (y-x)(14x+4y)) / (45x³)
    t1 = 45 * (x * x)
    t2 = y * (32 * (t1 - (y - x) * (14 * x + 4 * y)))
    t3 = x * t1
    atan_value = t2 // t3

    # Translate angles by sign flags

    cs.EUDSwitch(signflags, 7)
    swblock = ut.EUDGetLastBlock()[1]
    if cs.EUDSwitchCase()(0):  # xsign, ysign, xyabscmp = 0, 0, 0
        c.SetNextTrigger(swblock["swend"])
    if cs.EUDSwitchCase()(1):  # xsign, ysign, xyabscmp = 1, 0, 0
        c.RawTrigger(
            nextptr=swblock["swend"],
            actions=[
                atan_value.AddNumberX(0xFFFFFFFF, 0x55555555),
                atan_value.AddNumberX(0xFFFFFFFF, 0xAAAAAAAA),
                atan_value.AddNumber(128 + 1),
            ],
        )
    if cs.EUDSwitchCase()(2):  # xsign, ysign, xyabscmp = 0, 1, 0
        c.RawTrigger(
            nextptr=swblock["swend"],
            actions=[
                atan_value.AddNumberX(0xFFFFFFFF, 0x55555555),
                atan_value.AddNumberX(0xFFFFFFFF, 0xAAAAAAAA),
                atan_value.AddNumber(256 + 1),
            ],
        )
    if cs.EUDSwitchCase()(3):  # xsign, ysign, xyabscmp = 1, 1, 0
        c.RawTrigger(
            nextptr=swblock["swend"],
            actions=atan_value.AddNumber(128),
        )
    if cs.EUDSwitchCase()(4):  # xsign, ysign, xyabscmp = 0, 0, 1
        c.RawTrigger(
            nextptr=swblock["swend"],
            actions=[
                atan_value.AddNumberX(0xFFFFFFFF, 0x55555555),
                atan_value.AddNumberX(0xFFFFFFFF, 0xAAAAAAAA),
                atan_value.AddNumber(64 + 1),
            ],
        )
    if cs.EUDSwitchCase()(5):  # xsign, ysign, xyabscmp = 1, 0, 1
        c.RawTrigger(
            nextptr=swblock["swend"],
            actions=atan_value.AddNumber(64),
        )
    if cs.EUDSwitchCase()(6):  # xsign, ysign, xyabscmp = 0, 1, 1
        c.RawTrigger(
            nextptr=swblock["swend"],
            actions=atan_value.AddNumber(192),
        )
    if cs.EUDSwitchCase()(7):  # xsign, ysign, xyabscmp = 1, 1, 1
        c.RawTrigger(
            nextptr=swblock["swend"],
            actions=[
                atan_value.AddNumberX(0xFFFFFFFF, 0x55555555),
                atan_value.AddNumberX(0xFFFFFFFF, 0xAAAAAAAA),
                atan_value.AddNumber(192 + 1),
            ],
        )
    cs.EUDEndSwitch()
    c.RawTrigger(actions=atan_value.AddNumberX(64, 255))
    return atan_value



from helper import *


@TestInstance
def test_fmtprint():
    name = EUDVariable(Db("dpdkfah"))
    amount = EUDVariable(336)
    resource = EUDArray([EPD(Db("minerals")), EPD(Db("gases"))])
    f_setcurpl(P1)
    f_eprintAll("{:s} gets {} {:t}.", name, amount, resource[0])
    errorline = 0x640B60 + 218 * 12

    expect = "dpdkfah\r gets \r\r336\r \r\r\rminerals.\r\r\r"
    expectdb = Db(expect)
    test_equality("f_eprintf test1", f_strlen(errorline), len(expect))
    test_equality("f_eprintf test2", f_memcmp(errorline, expectdb, len(expect)), 0)



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from __future__ import annotations

from ... import utils as ut
from ...localize import _

"""
General CHK class.
"""


def sectionname_format(sn: str | bytes) -> bytes:
    section_name: bytes
    if isinstance(sn, str):
        section_name = sn.encode("ascii")
    else:
        section_name = sn

    if len(section_name) < 4:
        section_name += b" " * (4 - len(section_name))

    elif len(section_name) > 4:
        raise ut.EPError(_("Length of section name cannot be longer than 4"))

    return section_name


class CHK:
    def __init__(self) -> None:
        self.sections: dict[bytes, bytes] = {}

    def loadblank(self) -> None:
        self.sections = {}

    def loadchk(self, b: bytes) -> bool:
        # this code won't handle protection methods properly such as...
        # - duplicate section name
        # - jump section protection
        #
        # this program although handles
        #  - invalid section length (too high)
        #  - unused sections

        t = self.sections  # temporarily store
        self.sections = {}

        index = 0
        while index < len(b):
            # read data
            sectionname = b[index : index + 4]
            sectionlength = ut.b2i4(b, index + 4)

            if sectionlength < 0:
                # jsp with negative section size.
                self.sections = t
                return False

            section = b[index + 8 : index + 8 + sectionlength]
            index += sectionlength + 8

            self.sections[sectionname] = section

        return True

    def clone(self) -> CHK:
        t = CHK()
        t.sections = dict(self.sections)
        return t

    def savechk(self) -> bytes:
        # calculate output size
        blist = []
        for name, binary in self.sections.items():
            blist.append(name + ut.i2b4(len(binary)) + binary)

        fake_section = [b"ISOM"]
        import random

        fake_name = random.sample(fake_section, 1)
        blist.append(
            fake_name[0] + ut.i2b4(random.randint(0, 0xFFFFFFFF) | 0x80000000)
        )
        return b"".join(blist)

    def enumsection(self) -> list[bytes]:
        return list(self.sections.keys())

    def getsection(self, sectionname: str | bytes) -> bytes:
        sectionname = sectionname_format(sectionname)
        return self.sections[sectionname]  # KeyError may be raised.

    def setsection(self, sectionname: str | bytes, b: bytes) -> None:
        sectionname = sectionname_format(sectionname)
        self.sections[sectionname] = bytes(b)

    def delsection(self, sectionname: str | bytes) -> None:
        sectionname = sectionname_format(sectionname)
        del self.sections[sectionname]

    def optimize(self) -> None:
        # Delete unused sections
        # fmt: off
        used_section = [
            b'VER ', b'VCOD', b'OWNR', b'ERA ', b'DIM ', b'SIDE', b'MTXM',
            b'UNIT', b'THG2', b'MASK', b'STRx', b'UPRP', b'MRGN', b'TRIG',
            b'MBRF', b'SPRP', b'FORC', b'COLR', b'PUNI', b'PUPx', b'PTEx',
            b'UNIx', b'UPGx', b'TECx', b'CRGB',
        ]
        # fmt: on

        unused_section = [sn for sn in self.sections if sn not in used_section]
        for sn in unused_section:
            del self.sections[sn]

        # Omit MASK and MTXM optimization.
        # Game desync is reported if MTXM has shorter size.
        """
        # Terrain optimization
        dim = self.getsection(b"DIM ")
        mapw = ut.b2i2(dim, 0)
        maph = ut.b2i2(dim, 2)
        terrainsize = mapw * maph

        # MASK optimization : cancel 0xFFs.
        mask = self.getsection(b"MASK")
        if mask:
            clippos = 0
            for i in range(terrainsize - 1, -1, -1):
                if mask[i] != 0xFF:
                    clippos = i + 1
                    break

            mask = mask[:clippos]
            self.setsection(b"MASK", mask)

        # MTXM optimization
        # MASK optimization : cancel 0xFFs.
        mtxm = self.getsection(b"MTXM")
        clippos = 0
        for i in range(terrainsize - 1, -1, -1):
            if mtxm[2 * i] != 0x00 or mtxm[2 * i + 1] != 0x00:
                clippos = i + 1
                break

        mtxm = mtxm[: 2 * clippos]
        self.setsection(b"MTXM", mtxm)
        """

        # More optimization would be possible, but I don't care.



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from __future__ import annotations

from abc import ABCMeta, abstractmethod
from typing import TYPE_CHECKING, Literal, overload

from typing_extensions import Self

from .. import rawtrigger as bt
from ..allocator import ConstExpr, IsConstExpr

if TYPE_CHECKING:
    from ..rawtrigger.consttype import Dword


class VariableBase(metaclass=ABCMeta):
    __slots__ = ()

    @abstractmethod
    def __init__(self) -> None:
        pass

    @abstractmethod
    def getValueAddr(self) -> ConstExpr:  # noqa: N802
        raise NotImplementedError()

    # -------

    def AtLeast(self, value: Dword) -> bt.Condition:  # noqa: N802
        return bt.Memory(self.getValueAddr(), bt.AtLeast, value)

    def AtMost(self, value: Dword) -> bt.Condition:  # noqa: N802
        return bt.Memory(self.getValueAddr(), bt.AtMost, value)

    def Exactly(self, value: Dword) -> bt.Condition:  # noqa: N802
        return bt.Memory(self.getValueAddr(), bt.Exactly, value)

    # -------

    def SetNumber(self, value: Dword) -> bt.Action:  # noqa: N802
        return bt.SetMemory(self.getValueAddr(), bt.SetTo, value)

    def AddNumber(self, value: Dword) -> bt.Action:  # noqa: N802
        return bt.SetMemory(self.getValueAddr(), bt.Add, value)

    def SubtractNumber(self, value: Dword) -> bt.Action:  # noqa: N802
        return bt.SetMemory(self.getValueAddr(), bt.Subtract, value)

    # -------

    def AtLeastX(self, value: Dword, mask: Dword) -> bt.Condition:  # noqa: N802
        return bt.MemoryX(self.getValueAddr(), bt.AtLeast, value, mask)

    def AtMostX(self, value: Dword, mask: Dword) -> bt.Condition:  # noqa: N802
        return bt.MemoryX(self.getValueAddr(), bt.AtMost, value, mask)

    def ExactlyX(self, value: Dword, mask: Dword) -> bt.Condition:  # noqa: N802
        return bt.MemoryX(self.getValueAddr(), bt.Exactly, value, mask)

    # -------

    def SetNumberX(self, value: Dword, mask: Dword) -> bt.Action:  # noqa: N802
        return bt.SetMemoryX(self.getValueAddr(), bt.SetTo, value, mask)

    def AddNumberX(self, value: Dword, mask: Dword) -> bt.Action:  # noqa: N802
        return bt.SetMemoryX(self.getValueAddr(), bt.Add, value, mask)

    def SubtractNumberX(self, value: Dword, mask: Dword) -> bt.Action:  # noqa: N802
        return bt.SetMemoryX(self.getValueAddr(), bt.Subtract, value, mask)

    # -------

    def Assign(self: Self, value: Dword) -> Self:  # noqa: N802
        bt.RawTrigger(actions=bt.SetMemory(self.getValueAddr(), bt.SetTo, value))
        return self

    def __lshift__(self: Self, value: Dword) -> Self:
        return self.Assign(value)

    def __iadd__(self: Self, value: Dword) -> Self:
        bt.RawTrigger(actions=bt.SetMemory(self.getValueAddr(), bt.Add, value))
        return self

    def __isub__(self: Self, value: Dword) -> Self:
        bt.RawTrigger(actions=bt.SetMemory(self.getValueAddr(), bt.Subtract, value))
        return self

    # -------

    # See: https://github.com/heinermann/llvm-bw/wiki/Instruction-Implementation

    def __ior__(self: Self, value: Dword) -> Self:
        bt.RawTrigger(actions=self.SetNumberX(0xFFFFFFFF, value))
        return self

    def __iand__(self: Self, value: int | ConstExpr) -> Self:
        if isinstance(value, int):
            bt.RawTrigger(actions=self.SetNumberX(0, ~value))
        else:  # FIXME: calculate ~ConstExpr on compile time
            apply_and = self.SetNumberX(0, 0)
            bt.RawTrigger(
                actions=[
                    bt.SetMemory(apply_and, bt.SetTo, value),
                    bt.SetMemoryX(apply_and, bt.Add, 0xFFFFFFFF, 0x55555555),
                    bt.SetMemoryX(apply_and, bt.Add, 0xFFFFFFFF, 0xAAAAAAAA),
                    apply_and,
                ]
            )
        return self

    def __ixor__(self: Self, value: Dword) -> Self:
        bt.RawTrigger(
            actions=[
                self.AddNumberX(value, 0x55555555),  # 5 = 0b0101
                self.AddNumberX(value, 0xAAAAAAAA),  # A = 0b1010
            ]
        )
        return self

    def iinvert(self: Self) -> Self:
        "In-place invert (x << ~x)"
        return self.__ixor__(0xFFFFFFFF)

    @overload
    def ineg(self: Self, *, action: Literal[False] = False) -> Self:
        ...

    @overload
    def ineg(self: Self, *, action: Literal[True]) -> list[bt.Action]:
        ...

    def ineg(self: Self, *, action: bool = False) -> Self | list[bt.Action]:
        "In-place negate (x << -x)"
        actions = [
            self.AddNumberX(0xFFFFFFFF, 0x55555555),
            self.AddNumberX(0xFFFFFFFF, 0xAAAAAAAA),
            self.AddNumber(1),
        ]
        if action:
            return actions
        else:
            bt.RawTrigger(actions=actions)
            return self

    def iabs(self: Self) -> Self:
        bt.RawTrigger(
            conditions=self >= 0x80000000,
            actions=self.ineg(action=True),
        )
        return self

    def __ilshift__(self: Self, n: int) -> Self:  # type: ignore[misc]
        mask = (1 << (n + 1)) - 1
        actions = []
        for t in reversed(range(32 - n)):
            actions.append(self.SetNumberX(0, (mask >> 1) << (t + 1)))
            actions.append(self.AddNumberX((mask >> 1) << t, mask << t))
        actions.append(self.SetNumberX(0, mask >> 1))  # lowest n bits
        bt.RawTrigger(actions=actions)
        return self

    def __irshift__(self: Self, n: int) -> Self:
        mask = (1 << (n + 1)) - 1
        actions = [self.SetNumberX(0, mask >> 1)]  # lowest n bits
        actions.extend(
            self.SubtractNumberX((mask >> 1) << t, mask << t) for t in range(32 - n)
        )
        bt.RawTrigger(actions=actions)
        return self

    # -------

    def __eq__(self, other) -> bt.Condition:  # type: ignore[override]
        return self.Exactly(other)

    def __ne__(self, other) -> bt.Condition:  # type: ignore[override]
        if isinstance(other, int):
            if other & 0xFFFFFFFF == 0:
                return self.AtLeast(1)
            if other & 0xFFFFFFFF == 0xFFFFFFFF:
                return self.AtMost(0xFFFFFFFE)
        return NotImplemented

    def __le__(self, other: Dword) -> bt.Condition:
        return self.AtMost(other)

    def __lt__(self, other) -> bt.Condition:
        if IsConstExpr(other):
            return self.AtMost(other - 1)
        return NotImplemented

    def __ge__(self, other: Dword) -> bt.Condition:
        return self.AtLeast(other)

    def __gt__(self, other) -> bt.Condition:
        if IsConstExpr(other):
            return self.AtLeast(other + 1)
        return NotImplemented



# Copyright 2019 by Armoha.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from ... import core as c
from ... import ctrlstru as cs
from ... import utils as ut
from ...memio import f_dwread_epd


@c.EUDFunc
def f_getgametick():
    """Get current game tick value."""

    gametick_cache = c.EUDVariable()
    _gametick_cond = c.Forward()

    if cs.EUDIfNot()([_gametick_cond << c.Memory(0x57F23C, c.Exactly, 0)]):
        f_dwread_epd(ut.EPD(0x57F23C), ret=[gametick_cache])
        c.SetVariables(ut.EPD(_gametick_cond) + 2, gametick_cache)
    cs.EUDEndIf()

    return gametick_cache



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from .injector.apply_injector import PRT_SkipPayloadRelocator
from .injector.mainloop import EUDDoEvents, EUDOnStart
from .inlinecode.ilcprocesstrig import PRT_SetInliningRate
from .loadmap import LoadMap
from .mpqadd import MPQAddFile, MPQAddWave, MPQCheckFile
from .savemap import SaveMap

__all__ = [
    "PRT_SkipPayloadRelocator",
    "EUDDoEvents",
    "EUDOnStart",
    "PRT_SetInliningRate",
    "LoadMap",
    "MPQAddFile",
    "MPQAddWave",
    "MPQCheckFile",
    "SaveMap",
]



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from __future__ import annotations

import sys
import traceback
from collections.abc import Iterator, Sequence
from itertools import pairwise
from typing import TYPE_CHECKING, Any, TypeGuard

from typing_extensions import Self

from ...localize import _
from ...utils import (
    EPD,
    EPError,
    ExprProxy,
    FlattenList,
    List2Assignable,
    ep_assert,
    ep_warn,
    unProxy,
)
from .. import rawtrigger as bt
from ..allocator import ConstExpr, Forward
from .vbase import VariableBase
from .vbuf import get_current_varbuffer

if TYPE_CHECKING:
    from ..rawtrigger.constenc import TrgModifier
    from ..rawtrigger.consttype import Dword


_is_rvalue_strict = False


def EP_SetRValueStrictMode(mode: bool) -> None:  # noqa: N802
    global _is_rvalue_strict
    _is_rvalue_strict = mode


def process_dest(dest) -> int | ConstExpr:
    epd = unProxy(dest)
    if isinstance(epd, VariableBase):
        if isinstance(epd, EUDVariable):
            epd.checkNonRValue()
        return EPD(epd.getValueAddr())
    if not isinstance(epd, int | ConstExpr):
        raise EPError(_("Invalid dest: {}").format(dest))
    return epd


if sys.version_info >= (3, 11):
    _initial_refcount = 2
else:
    _initial_refcount = 3


def _is_rvalue(obj: object, refcount_bonus: int = 0, /) -> bool:
    refcount = _initial_refcount + 2 + refcount_bonus
    if isinstance(obj, EUDVariable) and sys.getrefcount(obj) != refcount:
        return False
    if isinstance(obj, ExprProxy) and not _is_rvalue(
        obj.getValue(), refcount_bonus + 1
    ):
        return False
    return True


def _yield_and_check_rvalue(
    obj: Any, refcount_bonus: int = 0, is_rvalue: bool = True
) -> Iterator[tuple[Any, bool]]:
    refcount = _initial_refcount + refcount_bonus
    is_rvalue &= sys.getrefcount(obj) <= refcount
    if isinstance(obj, ExprProxy):
        yield from _yield_and_check_rvalue(obj.getValue(), 1, is_rvalue)
    elif isinstance(obj, EUDVariable):
        yield obj, is_rvalue
    elif isinstance(obj, (bytes, str)) or hasattr(obj, "dont_flatten"):  # noqa: UP038
        yield obj, False
    else:
        try:
            for subobj in obj:
                yield from _yield_and_check_rvalue(subobj, 2, is_rvalue)
        except TypeError:  # obj is not iterable
            yield obj, False


# Unused variable don't need to be allocated.
class VariableTriggerForward(ConstExpr):
    def __new__(cls, *args, **kwargs) -> Self:
        return super().__new__(cls, None)

    def __init__(self, initval) -> None:
        super().__init__()
        self._initval = initval

    def Evaluate(self):  # noqa: N802
        evb = get_current_varbuffer()
        try:
            return evb._vdict[self].Evaluate()
        except KeyError:
            vt = evb.create_vartrigger(self, self._initval)
            return vt.Evaluate()


class EUDVariable(VariableBase):
    """Full variable."""

    _addor: EUDVariable
    __slots__ = ("_vartrigger", "_varact", "_rvalue")

    def __init__(self, initval=0) -> None:
        if not isinstance(initval, (int, ConstExpr)):  # noqa: UP038
            unproxied = unProxy(initval)
            if not isinstance(unproxied, (int, ConstExpr)):  # noqa: UP038
                raise EPError(_("Invalid initval: {}").format(initval))
            else:
                initval = unproxied
        self._vartrigger = VariableTriggerForward(initval)
        self._varact = self._vartrigger + (8 + 320)
        self._rvalue = False

    def GetVTable(self) -> ConstExpr:  # noqa: N802
        return self._vartrigger

    def getMaskAddr(self) -> ConstExpr:  # noqa: N802
        return self._varact

    def getDestAddr(self) -> ConstExpr:  # noqa: N802
        return self._varact + 16  # type: ignore[return-value]

    def getValueAddr(self) -> ConstExpr:  # noqa: N802
        return self._varact + 20  # type: ignore[return-value]

    def __hash__(self) -> int:
        return id(self)

    # -------
    def makeL(self) -> Self:  # noqa: N802
        self._rvalue = False
        return self

    def makeR(self) -> Self:  # noqa: N802
        self._rvalue = True
        return self

    def checkNonRValue(self) -> None:  # noqa: N802
        if _is_rvalue_strict and self._rvalue:
            raise EPError(_("Trying to modify value of r-value variable"))

    # -------

    def SetMask(self, value: Dword) -> bt.Action:  # noqa: N802
        return bt.SetMemory(self.getMaskAddr(), bt.SetTo, value)

    def AddMask(self, value: Dword) -> bt.Action:  # noqa: N802
        return bt.SetMemory(self.getMaskAddr(), bt.Add, value)

    def SubtractMask(self, value: Dword) -> bt.Action:  # noqa: N802
        return bt.SetMemory(self.getMaskAddr(), bt.Subtract, value)

    # -------

    def SetMaskX(self, value: Dword, mask: Dword) -> bt.Action:  # noqa: N802
        return bt.SetMemoryX(self.getMaskAddr(), bt.SetTo, value, mask)

    def AddMaskX(self, value: Dword, mask: Dword) -> bt.Action:  # noqa: N802
        return bt.SetMemoryX(self.getMaskAddr(), bt.Add, value, mask)

    def SubtractMaskX(self, value: Dword, mask: Dword) -> bt.Action:  # noqa: N802
        return bt.SetMemoryX(self.getMaskAddr(), bt.Subtract, value, mask)

    # -------

    def MaskAtLeast(self, value: Dword) -> bt.Condition:  # noqa: N802
        return bt.Memory(self.getMaskAddr(), bt.AtLeast, value)

    def MaskAtMost(self, value: Dword) -> bt.Condition:  # noqa: N802
        return bt.Memory(self.getMaskAddr(), bt.AtMost, value)

    def MaskExactly(self, value: Dword) -> bt.Condition:  # noqa: N802
        return bt.Memory(self.getMaskAddr(), bt.Exactly, value)

    # -------

    def MaskAtLeastX(self, value: Dword, mask: Dword) -> bt.Condition:  # noqa: N802
        return bt.MemoryX(self.getMaskAddr(), bt.AtLeast, value, mask)

    def MaskAtMostX(self, value: Dword, mask: Dword) -> bt.Condition:  # noqa: N802
        return bt.MemoryX(self.getMaskAddr(), bt.AtMost, value, mask)

    def MaskExactlyX(self, value: Dword, mask: Dword) -> bt.Condition:  # noqa: N802
        return bt.MemoryX(self.getMaskAddr(), bt.Exactly, value, mask)

    # -------

    def SetDest(self, dest) -> bt.Action:  # noqa: N802
        dest = process_dest(dest)
        return bt.SetMemory(self.getDestAddr(), bt.SetTo, dest)

    def AddDest(self, dest) -> bt.Action:  # noqa: N802
        dest = process_dest(dest)
        return bt.SetMemory(self.getDestAddr(), bt.Add, dest)

    def SubtractDest(self, dest) -> bt.Action:  # noqa: N802
        dest = process_dest(dest)
        return bt.SetMemory(self.getDestAddr(), bt.Subtract, dest)

    # -------

    def SetDestX(self, dest, mask: Dword) -> bt.Action:  # noqa: N802
        dest = process_dest(dest)
        return bt.SetMemoryX(self.getDestAddr(), bt.SetTo, dest, mask)

    def AddDestX(self, dest, mask: Dword) -> bt.Action:  # noqa: N802
        dest = process_dest(dest)
        return bt.SetMemoryX(self.getDestAddr(), bt.Add, dest, mask)

    def SubtractDestX(self, dest, mask: Dword) -> bt.Action:  # noqa: N802
        dest = process_dest(dest)
        return bt.SetMemoryX(self.getDestAddr(), bt.Subtract, dest, mask)

    # -------

    def SetModifier(self, modifier: TrgModifier) -> bt.Action:  # noqa: N802
        ep_assert(
            modifier is bt.SetTo or modifier is bt.Add or modifier is bt.Subtract,
            _("Unexpected modifier {}").format(modifier),
        )
        mode = bt.EncodeModifier(modifier) << 24
        return bt.SetDeathsX(EPD(self._varact + 24), bt.SetTo, mode, 0, 0xFF000000)

    # -------

    def QueueAssignTo(self, dest) -> list[bt.Action]:  # noqa: N802
        return [self.SetDest(dest), self.SetModifier(bt.SetTo)]

    def QueueAddTo(self, dest) -> list[bt.Action]:  # noqa: N802
        return [self.SetDest(dest), self.SetModifier(bt.Add)]

    def QueueSubtractTo(self, dest) -> list[bt.Action]:  # noqa: N802
        return [self.SetDest(dest), self.SetModifier(bt.Subtract)]

    # -------

    def Assign(self, other: Dword) -> Self:  # noqa: N802
        self.checkNonRValue()
        SeqCompute(((self, bt.SetTo, other),))
        return self

    def __lshift__(self, other: Dword) -> Self:
        return self.Assign(other)

    def __iadd__(self, other: Dword) -> Self:
        SeqCompute(((self, bt.Add, other),))
        return self

    def __isub__(self, other: Dword) -> Self:
        if isinstance(other, int):
            self.__iadd__(-other)  # 1A
        else:
            from .evcommon import _addor as addor

            SeqCompute(
                [  # (self + 1) + (~0 - other)
                    (self, bt.Add, 1),
                    (addor, bt.SetTo, 0xFFFFFFFF),
                    (addor, bt.Subtract, other),
                    (self, None, addor),
                ]
            )  # 1T 7A, executes 3T 9A
        return self

    # -------

    def __add__(self, other: Dword) -> EUDVariable:
        if _is_rvalue(self):
            return self.__iadd__(other)
        is_rvalue = _is_rvalue(other)
        unproxied = unProxy(other)
        if isinstance(unproxied, EUDVariable) and is_rvalue:
            return unproxied.__iadd__(self)
        t = EUDVariable()
        SeqCompute([(t, bt.SetTo, unproxied), (t, bt.Add, self)])
        return t.makeR()

    def __radd__(self, other: Dword) -> EUDVariable:
        if _is_rvalue(self):
            return self.__iadd__(other)
        t = EUDVariable()
        SeqCompute([(t, bt.SetTo, other), (t, bt.Add, self)])
        return t.makeR()

    def __sub__(self, other: Dword) -> EUDVariable:
        is_rvalue = _is_rvalue(other)
        unproxied = unProxy(other)
        if isinstance(unproxied, EUDVariable) and is_rvalue:
            rvalue_strict = _is_rvalue_strict
            if rvalue_strict:
                EP_SetRValueStrictMode(False)
            VProc(
                self,  # -other += self
                [
                    unproxied.AddNumberX(0xFFFFFFFF, 0x55555555),
                    unproxied.AddNumberX(0xFFFFFFFF, 0xAAAAAAAA),
                    unproxied.AddNumber(1),
                    self.QueueAddTo(unproxied),
                ],
            )  # 1T 7A, executes 2T 8A
            if rvalue_strict:
                EP_SetRValueStrictMode(True)
            return unproxied
        if _is_rvalue(self):
            return self.__isub__(unproxied)
        t = EUDVariable()
        # FIXME: unsupported EUD error after EUDStruct.free() with IsConstExpr
        if isinstance(unproxied, int):
            SeqCompute(
                [
                    (t, bt.SetTo, -unproxied),
                    (t, bt.Add, self),
                ]
            )
        else:
            SeqCompute(
                [
                    (t, bt.SetTo, 0xFFFFFFFF),
                    (t, bt.Subtract, unproxied),
                    (t, bt.Add, 1),
                    (t, bt.Add, self),
                ]
            )
        return t.makeR()

    def __rsub__(self, other: Dword) -> EUDVariable:
        unproxied = unProxy(other)
        is_eudvariable = isinstance(unproxied, EUDVariable)
        if _is_rvalue(self) and not is_eudvariable:
            bt.RawTrigger(  # -self += other
                actions=[
                    self.AddNumberX(0xFFFFFFFF, 0x55555555),
                    self.AddNumberX(0xFFFFFFFF, 0xAAAAAAAA),
                    self.AddNumber(unproxied + 1),
                ]
            )
            return self
        t = EUDVariable()
        if is_eudvariable:
            SeqCompute(
                [
                    (t, bt.SetTo, 0xFFFFFFFF),
                    (t, bt.Subtract, self),
                    (t, bt.Add, 1),
                    (t, bt.Add, unproxied),
                ]
            )
        else:
            SeqCompute(
                [
                    (t, bt.SetTo, 0xFFFFFFFF),
                    (t, bt.Subtract, self),
                    (t, bt.Add, unproxied + 1),
                ]
            )
        return t.makeR()

    def __neg__(self) -> EUDVariable:
        if _is_rvalue(self):
            return self.ineg()
        return (0 - self).makeR()

    def __invert__(self) -> EUDVariable:
        if _is_rvalue(self):
            return self.iinvert()
        t = EUDVariable()
        SeqCompute([(t, bt.SetTo, 0xFFFFFFFF), (t, bt.Subtract, self)])
        return t.makeR()

    # -------

    def __ior__(self, other: Dword) -> Self:
        unproxied = unProxy(other)
        if isinstance(unproxied, EUDVariable):
            write = self.SetNumberX(0xFFFFFFFF, 0)
            SeqCompute([(EPD(write), bt.SetTo, unproxied)])
            bt.RawTrigger(actions=write)  # 1T 5A
        else:
            super().__ior__(unproxied)  # 1A
        return self

    def __iand__(self, other: Dword) -> Self:
        unproxied = unProxy(other)
        if isinstance(unproxied, EUDVariable):
            write = self.SetNumberX(0, 0xFFFFFFFF)
            SeqCompute(
                [
                    (EPD(write), bt.SetTo, 0xFFFFFFFF),
                    (EPD(write), bt.Subtract, unproxied),
                ]
            )
            bt.RawTrigger(actions=write)  # 1T 6A
        else:
            super().__iand__(unproxied)  # 1A
        return self

    # -------

    def __and__(self, other: Dword) -> EUDVariable:
        if _is_rvalue(self):
            return self.__iand__(other)
        is_rvalue = _is_rvalue(other)
        unproxied = unProxy(other)
        if isinstance(unproxied, EUDVariable) and is_rvalue:
            return unproxied.__iand__(self)
        t = EUDVariable()
        if isinstance(unproxied, EUDVariable):
            write = t.SetNumberX(0, 0xFFFFFFFF)
            SeqCompute(
                [
                    (EPD(write), bt.SetTo, 0xFFFFFFFF),
                    (t, bt.SetTo, self),
                    (EPD(write), bt.Subtract, unproxied),
                ]
            )
            bt.RawTrigger(actions=write)  # 2T 10A
        else:
            t << self
            t &= unproxied  # 1T 5A
        return t.makeR()

    def __rand__(self, other: Dword) -> EUDVariable:
        if _is_rvalue(self):
            return self.__iand__(other)
        t = EUDVariable()
        unproxied = unProxy(other)
        if isinstance(unproxied, EUDVariable):
            write = t.SetNumberX(0, 0xFFFFFFFF)
            SeqCompute(
                [
                    (EPD(write), bt.SetTo, 0xFFFFFFFF),
                    (t, bt.SetTo, self),
                    (EPD(write), bt.Subtract, unproxied),
                ]
            )
            bt.RawTrigger(actions=write)  # 2T 10A
        else:
            t << self
            t &= unproxied  # 1T 5A
        return t.makeR()

    def __or__(self, other: Dword) -> EUDVariable:
        if _is_rvalue(self):
            return self.__ior__(other)
        is_rvalue = _is_rvalue(other)
        unproxied = unProxy(other)
        if isinstance(unproxied, EUDVariable) and is_rvalue:
            return unproxied.__ior__(self)
        t = EUDVariable()
        if isinstance(unproxied, EUDVariable):
            write = t.SetNumberX(0xFFFFFFFF, 0)
            SeqCompute([(t, bt.SetTo, self), (EPD(write), bt.SetTo, unproxied)])
            bt.RawTrigger(actions=write)  # 2T 9A
        else:
            t << self
            t |= unproxied  # 1T 5A
        return t.makeR()

    def __ror__(self, other: Dword) -> EUDVariable:
        if _is_rvalue(self):
            return self.__ior__(other)
        t = EUDVariable()
        unproxied = unProxy(other)
        if isinstance(unproxied, EUDVariable):
            write = t.SetNumberX(0xFFFFFFFF, 0)
            SeqCompute([(t, bt.SetTo, self), (EPD(write), bt.SetTo, unproxied)])
            bt.RawTrigger(actions=write)  # 2T 9A
        else:
            t << self
            t |= unproxied  # 1T 5A
        return t.makeR()

    # -------

    def __ixor__(self, other: Dword) -> Self:
        unproxied = unProxy(other)
        if isinstance(unproxied, EUDVariable):
            VProc(
                unproxied,
                [
                    unproxied.SetMask(0x55555555),
                    *unproxied.QueueAddTo(self),
                ],
            )
            VProc(unproxied, unproxied.SetMask(0xAAAAAAAA))  # 3T 8A
            bt.RawTrigger(
                actions=unproxied.SetMask(0xFFFFFFFF)
            )  # FIXME: restore to previous mask???
        else:
            super().__ixor__(unproxied)  # 2A
        return self

    def __xor__(self, other: Dword) -> EUDVariable:
        if _is_rvalue(self):
            return self.__ixor__(other)
        is_rvalue = _is_rvalue(other)
        unproxied = unProxy(other)
        if isinstance(unproxied, EUDVariable) and is_rvalue:
            return unproxied.__ixor__(self)
        t = EUDVariable()
        if isinstance(unproxied, EUDVariable):
            VProc(
                [self, unproxied],
                [
                    unproxied.SetMask(0x55555555),
                    self.QueueAssignTo(t),
                    unproxied.QueueAddTo(t),
                ],
            )
            VProc(unproxied, unproxied.SetMask(0xAAAAAAAA))
            bt.RawTrigger(
                actions=unproxied.SetMask(0xFFFFFFFF)
            )  # FIXME: restore to previous mask???
        else:
            t << self
            t ^= unproxied
        return t.makeR()

    def __rxor__(self, other: Dword) -> EUDVariable:
        unproxied = unProxy(other)
        if _is_rvalue(self):
            return self.__ixor__(unproxied)
        t = EUDVariable()
        if isinstance(unproxied, EUDVariable):
            SeqCompute(
                [
                    (EPD(unproxied.getMaskAddr()), bt.SetTo, 0x55555555),
                    (t, bt.SetTo, self),
                    (t, bt.Add, unproxied),
                ]
            )
            VProc(unproxied, unproxied.SetMask(0xAAAAAAAA))
            bt.RawTrigger(
                actions=unproxied.SetMask(0xFFFFFFFF)
            )  # FIXME: restore to previous mask???
        else:
            t << self
            t ^= unproxied
        return t.makeR()

    # -------

    def __eq__(self, other) -> bt.Condition:  # type: ignore[override]
        try:
            return self.Exactly(other)

        except Exception as err:
            ep_warn(_("{}: Comparing with temporary variable.").format(err))
            traceback.print_stack()
            return (self - other).Exactly(0)

    def __ne__(self, other) -> bt.Condition:  # type: ignore[override]
        if isinstance(other, int):
            if other & 0xFFFFFFFF == 0:
                return self.AtLeast(1)
            if other & 0xFFFFFFFF == 0xFFFFFFFF:
                return self.AtMost(0xFFFFFFFE)
        return (self - other).AtLeast(1)

    def __le__(self, other) -> bt.Condition:  # type: ignore[override]
        try:
            return self.AtMost(other)

        except Exception as err:
            ep_warn(_("{}: Patching comparison condition.").format(err))
            traceback.print_stack()
            condition = self.AtMost(0)
            SeqCompute(((EPD(condition) + 2, bt.SetTo, other),))
            return condition

    def __ge__(self, other) -> bt.Condition:  # type: ignore[override]
        try:
            return self.AtLeast(other)

        except Exception as err:
            ep_warn(_("{}: Patching comparison condition.").format(err))
            traceback.print_stack()
            condition = self.AtLeast(0)
            SeqCompute(((EPD(condition) + 2, bt.SetTo, other),))
            return condition

    def __lt__(self, other) -> bt.Condition:  # type: ignore[override]
        other = unProxy(other)
        if isinstance(other, EUDVariable):
            bitmask = Forward()  # u32: Location number or bitmask for the condition
            condition = bt.MemoryEPD(bitmask, bt.AtLeast, 1)
            bitmask << EPD(condition)
            SeqCompute(((bitmask, bt.SetTo, other), (bitmask, bt.Subtract, self)))
            return condition
        if isinstance(other, int) and other <= 0:
            ep_warn(_("No unsigned number can be leq than {}").format(other))
            traceback.print_stack()
            return bt.Never()  # No unsigned number is less than 0
        return self.AtMost(other - 1)

    def __gt__(self, other) -> bt.Condition:  # type: ignore[override]
        other = unProxy(other)
        if isinstance(other, EUDVariable):
            bitmask = Forward()  # u32: Location number or bitmask for the condition
            condition = bt.MemoryEPD(bitmask, bt.AtLeast, 1)
            bitmask << EPD(condition)
            SeqCompute(((bitmask, bt.SetTo, self), (bitmask, bt.Subtract, other)))
            return condition
        if isinstance(other, int) and other >= 0xFFFFFFFF:
            ep_warn(_("No unsigned number can be greater than {}").format(other))
            traceback.print_stack()
            return bt.Never()  # No unsigned number is greater than 0xFFFFFFFF
        return self.AtLeast(other + 1)

    # operator placeholders
    def __mul__(self, a):
        raise NotImplementedError

    def __rmul__(self, a):
        raise NotImplementedError

    def __imul__(self, a):
        raise NotImplementedError

    def __floordiv__(self, a):
        raise NotImplementedError

    def __rfloordiv__(self, a):
        raise NotImplementedError

    def __ifloordiv__(self, a):
        raise NotImplementedError

    def __mod__(self, a):
        raise NotImplementedError

    def __rmod__(self, a):
        raise NotImplementedError

    def __imod__(self, a):
        raise NotImplementedError

    def __ilshift__(self, a):  # type: ignore[misc]
        raise NotImplementedError

    def __irshift__(self, a):
        raise NotImplementedError


def IsEUDVariable(x: object) -> TypeGuard[EUDVariable]:  # noqa: N802
    return isinstance(unProxy(x), EUDVariable)


# ---------


def VProc(  # noqa: N802
    v: EUDVariable | Sequence[EUDVariable], actions
) -> bt.RawTrigger:
    v = FlattenList(v)
    actions = FlattenList(actions)
    end = Forward()
    first_trigger = None

    for cv, nv in pairwise(v):
        actions.append(bt.SetNextPtr(cv.GetVTable(), nv.GetVTable()))
    actions.append(bt.SetNextPtr(v[-1].GetVTable(), end))

    for i in range(0, len(actions), 64):
        trigger = bt.RawTrigger(actions=actions[i : i + 64])
        if first_trigger is None:
            first_trigger = trigger
    bt.SetNextTrigger(v[0].GetVTable())
    end << bt.NextTrigger()

    return first_trigger  # type: ignore[return-value]


# From vbuffer.py
def EUDCreateVariables(varn: int):  # noqa: N802
    return List2Assignable([EUDVariable() for _ in range(varn)])


# -------


def _get_computedest(dst):
    try:
        return EPD(dst.getValueAddr())
    except AttributeError:
        return dst


def _seqcompute_sub(assignpairs, _srcdict):
    """
    Subset of SeqCompute with following restrictions

    - Assignment from variable should be after assignment from constant.
    - Total number of actions should be leq than 64
    """

    actionlist = []

    # Collect constant-assigning actions
    const_assigning_index = len(assignpairs)

    for i, assignpair in enumerate(assignpairs):
        dst, mdt, src = assignpair
        if IsEUDVariable(src):
            const_assigning_index = i
            break

    for dst, mdt, src in assignpairs[0:const_assigning_index]:
        dst = _get_computedest(dst)
        actionlist.append(bt.SetDeaths(dst, mdt, src, 0))

    # Only constant-assigning function : skip
    if const_assigning_index == len(assignpairs):
        bt.RawTrigger(actions=actionlist)
        return

    #
    # Rest is for non-constant assigning actions
    #
    nextptr = None  # nextptr for this rawtrigger
    vt_nextptr = None  # what to set for nextptr of current vtable
    last_pairs = None
    non_const_actions = []

    def remove_duplicate_actions() -> None:
        if last_pairs is None:
            return
        src, dst, mdt = last_pairs

        try:
            prev_dst, prev_mdt, prev_nptr = _srcdict[src]
        except KeyError:
            prev_dst, prev_mdt, prev_nptr = None, None, None

        if (dst is not prev_dst) and dst:
            non_const_actions.append(src.SetDest(dst))
        if (mdt is not prev_mdt) and mdt:
            non_const_actions.append(src.SetModifier(mdt))
        if prev_nptr is None or vt_nextptr.expr is not prev_nptr.expr:
            non_const_actions.append(bt.SetNextPtr(src.GetVTable(), vt_nextptr))

        _srcdict[src] = (
            dst if dst else prev_dst,
            mdt if mdt else prev_mdt,
            vt_nextptr,
        )

    for dst, mdt, src in assignpairs[const_assigning_index:]:
        dst = _get_computedest(dst)

        if nextptr is None:
            nextptr = src.GetVTable()
        else:
            vt_nextptr << src.GetVTable()

        remove_duplicate_actions()

        vt_nextptr = Forward()
        last_pairs = src, dst, mdt

    remove_duplicate_actions()
    actionlist.extend(non_const_actions)
    bt.RawTrigger(nextptr=nextptr, actions=actionlist)

    vt_nextptr << bt.NextTrigger()


def SeqCompute(assignpairs):  # noqa: N802
    # We need dependency map while writing assignment pairs
    dstvarset = set()
    srcvarset = set()

    # Record previous dst, mdt for src to optimize duplicate actions
    srcdictsub = {}
    srcdict = {}

    # Sublist of assignments to put in _seqcompute_sub
    subassignpairs = []

    # Is we collecting constant-assigning pairs?
    constcollecting = True

    # Number of expected actions.
    actioncount = 0

    def flush_pairs():
        nonlocal constcollecting, actioncount

        if actioncount == 0:  # Already flushed before
            return

        _seqcompute_sub(subassignpairs, srcdictsub)

        dstvarset.clear()
        srcvarset.clear()
        subassignpairs.clear()
        constcollecting = True
        actioncount = 0

    for assignpair in assignpairs:
        dst, mdt, src = assignpair
        dst = bt.EncodePlayer(unProxy(dst))
        src = unProxy(src)

        # Flush action set before proceeding
        if IsEUDVariable(src):
            if src in srcvarset:
                flush_pairs()
            elif actioncount >= 64 - 3:
                flush_pairs()

            srcvarset.add(src)
            constcollecting = False
            actioncount += 3

            try:
                prev_dst, prev_mdt = srcdict[src]
            except KeyError:
                pass
            else:
                if dst is prev_dst:
                    actioncount -= 1
                if mdt is prev_mdt:
                    actioncount -= 1
            srcdict[src] = dst, mdt

        else:
            if not constcollecting:
                flush_pairs()
            elif actioncount >= 64 - 3:
                flush_pairs()

            actioncount += 1

        subassignpairs.append((dst, mdt, src))
        if IsEUDVariable(dst):
            dstvarset.add(dst)

    flush_pairs()


def NonSeqCompute(assignpairs):  # noqa: N802
    import itertools

    dstvarset = set()
    srcvarset = set()
    constpairs = list()
    varassigndict = dict()

    for assignpair in assignpairs:
        dst, mdt, src = assignpair
        if IsEUDVariable(dst):
            dstvarset.add(dst)
        if IsEUDVariable(src):
            srcvarset.add(src)
            try:
                varassigndict[src].append(assignpair)
            except KeyError:
                varassigndict[src] = [assignpair]
        else:
            constpairs.append(assignpair)

    if len(assignpairs) == len(constpairs):
        SeqCompute(assignpairs)
        return
    ep_assert(dstvarset.isdisjoint(srcvarset), _("dst and src have intersection"))

    varpairlists = list()
    for pairlist in varassigndict.values():
        pairlist.sort(key=lambda x: id(x[0]))
        pairlist.sort(key=lambda x: getattr(x[1], "_name", ""))
        varpairlists.append(pairlist)

    varassignpairs = list()
    for vp in itertools.zip_longest(*varpairlists):
        for varassignpair in vp:
            if varassignpair is not None:
                varassignpairs.append(varassignpair)

    SeqCompute(constpairs + varassignpairs)


def SetVariables(srclist, dstlist, mdtlist=None) -> None:  # noqa: N802
    srclist_refcount = _initial_refcount
    errlist = []
    if sys.getrefcount(srclist) == srclist_refcount:
        nth = 0
        for src, is_rvalue in _yield_and_check_rvalue(srclist, 1):
            if isinstance(src, EUDVariable) and is_rvalue:
                errlist.append(EPError(_("src{} is RValue variable").format(nth)))
            nth += 1
    srclist = FlattenList(srclist)
    dstlist = FlattenList(dstlist)
    if len(srclist) != len(dstlist):
        errlist.append(EPError(_("Input/output size mismatch")))
    if len(errlist) == 1:
        raise errlist[0]
    elif errlist:
        if sys.version_info >= (3, 11):
            raise ExceptionGroup(
                _("Multiple error occurred on SetVariables:"), errlist
            )
        else:
            raise EPError(_("Multiple error occurred on SetVariables:"), errlist)

    if mdtlist is None:
        mdtlist = [bt.SetTo] * len(srclist)
    else:
        mdtlist = FlattenList(mdtlist)

    sqa = [(src, mdt, dst) for src, dst, mdt in zip(srclist, dstlist, mdtlist)]
    SeqCompute(sqa)



# Copyright 2019 by Armoha.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from .. import core as c
from .. import ctrlstru as cs
from .. import utils as ut
from ..core.allocator.payload import _register_after_collecting_callback
from ..core.curpl import _curpl_checkcond, _curpl_var
from ..core.mapdata.stringmap import get_string_section_name
from ..localize import _
from ..memio import f_dwread_epd, f_wread_epd
from ..scdata import CurrentPlayer

_STR_ADDRESS = 0x191943C8


@c.EUDTypedFunc([c.TrgString])
def _get_mapstring_addr(str_id):
    add_strptr, add_strepd = c.Forward(), c.Forward()
    if cs.EUDExecuteOnce()():
        strptr, strepd = _STR_ADDRESS, ut.EPD(_STR_ADDRESS)
        cs.DoActions(
            c.SetMemory(add_strptr + 20, c.SetTo, strptr),
            c.SetMemory(add_strepd + 20, c.SetTo, strepd),
        )
    cs.EUDEndExecuteOnce()
    str_chunk_name = get_string_section_name()
    if str_chunk_name == "STR":
        r, m = c.f_div(str_id, 2)
        c.RawTrigger(conditions=m.Exactly(1), actions=m.SetNumber(2))
        c.RawTrigger(actions=add_strepd << r.AddNumber(0))
        ret = f_wread_epd(r, m)
    elif str_chunk_name == "STRx":
        c.RawTrigger(actions=add_strepd << str_id.AddNumber(0))
        ret = f_dwread_epd(str_id)
    else:
        raise ut.EPError(_("Invalid string section name: {}").format(str_chunk_name))
    c.RawTrigger(actions=add_strptr << ret.AddNumber(0))
    c.EUDReturn(ret)


_const_strptr: dict[int, c.Forward] = {}


def _initialize_queries():
    from ..core.mapdata.stringmap import get_string_map

    strmap = get_string_map()
    strmap.finalize()
    non_existing_id = []
    for str_id, offset_query in _const_strptr.items():
        try:
            offset_query.Reset()
            offset_query << c.ConstExpr(
                None,
                _STR_ADDRESS + strmap._stroffset[strmap._dataindextb[str_id - 1]],
                0,
            )
        except IndexError:
            non_existing_id.append(str_id)

    if non_existing_id:
        raise ut.EPError(
            _("GetMapStringAddr(str_id) for non-existing string ID(s): {}").format(
                non_existing_id
            )
        )


_register_after_collecting_callback(_initialize_queries)


def GetMapStringAddr(str_id):  # noqa: N802
    global _const_strptr
    str_id = c.EncodeString(str_id)
    if isinstance(str_id, int):
        if str_id not in _const_strptr:
            _const_strptr[str_id] = c.Forward()
            _const_strptr[str_id] << 0
        return _const_strptr[str_id]
    return _get_mapstring_addr(str_id)


def _s2b(x):
    if isinstance(x, str):
        x = ut.u2utf8(x)
    if isinstance(x, bytes):
        x = x + b"\r" * (-(-len(x) // 4) * 4 - len(x))
    return x


def _addcpcache(p):
    p = c.EncodePlayer(p)
    return [
        _curpl_var.AddNumber(p),
        c.SetMemory(_curpl_checkcond + 8, c.Add, p),
    ]


class CPString:
    """
    store String in SetDeaths Actions, easy to concatenate.
    """

    def __init__(self, content=None):
        """Constructor for CPString
        :param content: Initial CPString content / capacity. Capacity of
            CPString is determined by size of this. If content is integer, then
            initial capacity and content of CPString will be set to
            content(int) and empty string.
        :type content: str, bytes, int
        """
        if isinstance(content, int):
            self.content = b"\r" * -(-content // 4) * 4
        elif isinstance(content, str) or isinstance(content, bytes):
            self.content = _s2b(content)
        else:
            raise ut.EPError(
                _("Unexpected type for CPString: {}").format(type(content))
            )

        self.length = len(self.content) // 4
        self.trigger = list()
        self.valueAddr = [0 for _ in range(self.length)]
        actions = [
            [
                c.SetDeaths(
                    CurrentPlayer,
                    c.SetTo,
                    ut.b2i4(self.content[i : i + 4]),
                    i // 48,
                )
                for i in range(4 * mod, len(self.content), 48)
            ]
            for mod in range(12)
        ]
        modlength = self.length
        addr = 0
        for i, a in enumerate(actions):
            for k in range(len(a)):
                self.valueAddr[i + 12 * k] = addr
                addr += 1
            if a and i < 11:
                a.append(c.SetMemory(0x6509B0, c.Add, 1))
                addr += 1
                modlength -= 1
        actions.extend(
            [
                c.SetMemory(0x6509B0, c.Add, modlength) if modlength else [],
                _addcpcache(self.length),
            ]
        )
        actions = ut.FlattenList(actions)
        c.PushTriggerScope()
        for i in range(0, len(actions), 64):
            t = c.RawTrigger(actions=actions[i : i + 64])
            self.trigger.append(t)
        c.PopTriggerScope()

        self.valueAddr = [
            self.trigger[v // 64] + 348 + 32 * (v % 64) for v in self.valueAddr
        ]
        _nextptr = c.Forward()
        self.trigger[-1]._nextptr = _nextptr
        _nextptr << c.NextTrigger()

    def Display(self, action=[]):  # noqa: N802
        _next = c.Forward()
        c.RawTrigger(
            nextptr=self.trigger[0],
            actions=[action, c.SetNextPtr(self.trigger[-1], _next)],
        )
        _next << c.NextTrigger()

    def GetVTable(self):  # noqa: N802
        return self.trigger[0]



from typing import Literal, TypeAlias

DefButtonSetDict: dict[str, int] = {
    "Terran Marine": 0,
    "Terran Ghost": 1,
    "Terran Vulture": 2,
    "Terran Goliath": 3,
    "Goliath Turret": 4,
    "Terran Siege Tank (Tank Mode)": 5,
    "Siege Tank Turret (Tank Mode)": 6,
    "Tank Turret type   1": 6,
    "Terran SCV": 7,
    "Terran Wraith": 8,
    "Terran Science Vessel": 9,
    "Gui Montag": 10,
    "Gui Montag (Firebat)": 10,
    "Terran Dropship": 11,
    "Terran Battlecruiser": 12,
    "Spider Mine": 13,
    "Vulture Spider Mine": 13,
    "Nuclear Missile": 14,
    "Terran Civilian": 15,
    "Sarah Kerrigan": 16,
    "Sarah Kerrigan (Ghost)": 16,
    "Alan Schezar": 17,
    "Alan Schezar (Goliath)": 17,
    "Alan Schezar Turret": 18,
    "Alan Turret": 18,
    "Jim Raynor (Vulture)": 19,
    "Jim Raynor (Marine)": 20,
    "Tom Kazansky": 21,
    "Tom Kazansky (Wraith)": 21,
    "Magellan": 22,
    "Magellan (Science Vessel)": 22,
    "Edmund Duke (Tank Mode)": 23,
    "Edmund Duke Turret (Tank Mode)": 24,
    "Duke Turret type   1": 24,
    "Edmund Duke (Siege Mode)": 25,
    "Edmund Duke Turret (Siege Mode)": 26,
    "Duke Turret type   2": 26,
    "Arcturus Mengsk": 27,
    "Arcturus Mengsk (Battlecruiser)": 27,
    "Hyperion": 28,
    "Hyperion (Battlecruiser)": 28,
    "Norad II (Battlecruiser)": 29,
    "Terran Siege Tank (Siege Mode)": 30,
    "Siege Tank Turret (Siege Mode)": 31,
    "Tank Turret type   2": 31,
    "Terran Firebat": 32,
    "Scanner Sweep": 33,
    "Terran Medic": 34,
    "Zerg Larva": 35,
    "Zerg Egg": 36,
    "Zerg Zergling": 37,
    "Zerg Hydralisk": 38,
    "Zerg Ultralisk": 39,
    "Zerg Broodling": 40,
    "Zerg Drone": 41,
    "Zerg Overlord": 42,
    "Zerg Mutalisk": 43,
    "Zerg Guardian": 44,
    "Zerg Queen": 45,
    "Zerg Defiler": 46,
    "Zerg Scourge": 47,
    "Torrasque": 48,
    "Torrasque (Ultralisk)": 48,
    "Matriarch": 49,
    "Matriarch (Queen)": 49,
    "Infested Terran": 50,
    "Infested Kerrigan": 51,
    "Infested Kerrigan (Infested Terran)": 51,
    "Unclean One": 52,
    "Unclean One (Defiler)": 52,
    "Hunter Killer": 53,
    "Hunter Killer (Hydralisk)": 53,
    "Devouring One": 54,
    "Devouring One (Zergling)": 54,
    "Kukulza (Mutalisk)": 55,
    "Kukulza (Guardian)": 56,
    "Yggdrasill": 57,
    "Yggdrasill (Overlord)": 57,
    "Terran Valkyrie": 58,
    "Mutalisk Cocoon": 59,
    "Cocoon": 59,
    "Protoss Corsair": 60,
    "Protoss Dark Templar": 61,
    "Protoss Dark Templar (Unit)": 61,
    "Zerg Devourer": 62,
    "Protoss Dark Archon": 63,
    "Protoss Probe": 64,
    "Protoss Zealot": 65,
    "Protoss Dragoon": 66,
    "Protoss High Templar": 67,
    "Protoss Archon": 68,
    "Protoss Shuttle": 69,
    "Protoss Scout": 70,
    "Protoss Arbiter": 71,
    "Protoss Carrier": 72,
    "Protoss Interceptor": 73,
    "Dark Templar": 74,
    "Dark Templar (Hero)": 74,
    "Protoss Dark Templar (Hero)": 74,
    "Zeratul": 75,
    "Zeratul (Dark Templar)": 75,
    "Tassadar/Zeratul": 76,
    "Tassadar/Zeratul (Archon)": 76,
    "Fenix (Zealot)": 77,
    "Fenix (Dragoon)": 78,
    "Tassadar": 79,
    "Tassadar (Templar)": 79,
    "Mojo": 80,
    "Mojo (Scout)": 80,
    "Warbringer": 81,
    "Warbringer (Reaver)": 81,
    "Gantrithor": 82,
    "Gantrithor (Carrier)": 82,
    "Protoss Reaver": 83,
    "Protoss Observer": 84,
    "Protoss Scarab": 85,
    "Danimoth": 86,
    "Danimoth (Arbiter)": 86,
    "Aldaris": 87,
    "Aldaris (Templar)": 87,
    "Artanis": 88,
    "Artanis (Scout)": 88,
    "Rhynadon": 89,
    "Rhynadon (Badlands)": 89,
    "Rhynadon (Badlands Critter)": 89,
    "Bengalaas": 90,
    "Bengalaas (Jungle)": 90,
    "Bengalaas (Jungle Critter)": 90,
    "Cargo Ship": 91,
    "Cargo Ship (Unused)": 91,
    "Unused type   1": 91,
    "Unused type   2": 92,
    "Mercenary Gunship": 92,
    "Mercenary Gunship (Unused)": 92,
    "Scantid": 93,
    "Scantid (Desert)": 93,
    "Scantid (Desert Critter)": 93,
    "Kakaru": 94,
    "Kakaru (Twilight)": 94,
    "Kakaru (Twilight Critter)": 94,
    "Ragnasaur": 95,
    "Ragnasaur (Ashworld)": 95,
    "Ragnasaur (Ashworld Critter)": 95,
    "Ursadon": 96,
    "Ursadon (Ice World)": 96,
    "Ursadon (Ice World Critter)": 96,
    "Lurker Egg": 97,
    "Zerg Lurker Egg": 97,
    "Raszagal": 98,
    "Raszagal (Corsair)": 98,
    "Samir Duran": 99,
    "Samir Duran (Ghost)": 99,
    "Alexei Stukov": 100,
    "Alexei Stukov (Ghost)": 100,
    "Map Revealer": 101,
    "Gerard DuGalle": 102,
    "Gerard DuGalle (Ghost)": 102,
    "Gerard DuGalle (BattleCruiser)": 102,
    "Zerg Lurker": 103,
    "Infested Duran": 104,
    "Infested Duran (Infested Terran)": 104,
    "Disruption Web": 105,
    "Disruption Field": 105,
    "Terran Command Center": 106,
    "Terran Comsat Station": 107,
    "Terran Nuclear Silo": 108,
    "Terran Supply Depot": 109,
    "Terran Refinery": 110,
    "Terran Barracks": 111,
    "Terran Academy": 112,
    "Terran Factory": 113,
    "Terran Starport": 114,
    "Terran Control Tower": 115,
    "Terran Science Facility": 116,
    "Terran Covert Ops": 117,
    "Terran Physics Lab": 118,
    "Starbase": 119,
    "Starbase (Unused)": 119,
    "Unused Terran Bldg type   1": 119,
    "Terran Machine Shop": 120,
    "Repair Bay": 121,
    "Repair Bay (Unused)": 121,
    "Unused Terran Bldg type   2": 121,
    "Terran Engineering Bay": 122,
    "Terran Armory": 123,
    "Terran Missile Turret": 124,
    "Terran Bunker": 125,
    "Norad II (Crashed)": 126,
    "Norad II (Crashed Battlecruiser)": 126,
    "Ion Cannon": 127,
    "Uraj Crystal": 128,
    "Khalis Crystal": 129,
    "Infested Command Center": 130,
    "Zerg Hatchery": 131,
    "Zerg Lair": 132,
    "Zerg Hive": 133,
    "Zerg Nydus Canal": 134,
    "Zerg Hydralisk Den": 135,
    "Zerg Defiler Mound": 136,
    "Zerg Greater Spire": 137,
    "Zerg Queen's Nest": 138,
    "Zerg Evolution Chamber": 139,
    "Zerg Ultralisk Cavern": 140,
    "Zerg Spire": 141,
    "Zerg Spawning Pool": 142,
    "Zerg Creep Colony": 143,
    "Zerg Spore Colony": 144,
    "Unused Zerg Bldg": 145,
    "Unused Zerg Building1": 145,
    "Zerg Sunken Colony": 146,
    "Zerg Overmind (With Shell)": 147,
    "Zerg Overmind": 148,
    "Zerg Extractor": 149,
    "Mature Chrysalis": 150,
    "Zerg Cerebrate": 151,
    "Zerg Cerebrate Daggoth": 152,
    "Unused Zerg Building2": 153,
    "Protoss Nexus": 154,
    "Protoss Robotics Facility": 155,
    "Protoss Pylon": 156,
    "Protoss Assimilator": 157,
    "Protoss Unused type   1": 158,
    "Unused Protoss Building1": 158,
    "Protoss Observatory": 159,
    "Protoss Gateway": 160,
    "Protoss Unused type   2": 161,
    "Unused Protoss Building2": 161,
    "Protoss Photon Cannon": 162,
    "Protoss Citadel of Adun": 163,
    "Protoss Cybernetics Core": 164,
    "Protoss Templar Archives": 165,
    "Protoss Forge": 166,
    "Protoss Stargate": 167,
    "Stasis Cell/Prison": 168,
    "Protoss Fleet Beacon": 169,
    "Protoss Arbiter Tribunal": 170,
    "Protoss Robotics Support Bay": 171,
    "Protoss Shield Battery": 172,
    "Khaydarin Crystal Formation": 173,
    "Protoss Temple": 174,
    "Xel'Naga Temple": 175,
    "Mineral Field (Type 1)": 176,
    "Mineral Field (Type 2)": 177,
    "Mineral Field (Type 3)": 178,
    "Cave": 179,
    "Cave (Unused)": 179,
    "Cave-in": 180,
    "Cave-in (Unused)": 180,
    "Cantina": 181,
    "Cantina (Unused)": 181,
    "Mining Platform": 182,
    "Mining Platform (Unused)": 182,
    "Independent Command Center": 183,
    "Independent Command Center (Unused)": 183,
    "Independent Starport": 184,
    "Independent Starport (Unused)": 184,
    "Jump Gate": 185,
    "Independent Jump Gate (Unused)": 185,
    "Ruins": 186,
    "Ruins (Unused)": 186,
    "Kyadarin Crystal Formation": 187,
    "Khaydarin Crystal Formation (Unused)": 187,
    "Vespene Geyser": 188,
    "Warp Gate": 189,
    "Psi Disrupter": 190,
    "Zerg Marker": 191,
    "Terran Marker": 192,
    "Protoss Marker": 193,
    "Zerg Beacon": 194,
    "Terran Beacon": 195,
    "Protoss Beacon": 196,
    "Zerg Flag Beacon": 197,
    "Terran Flag Beacon": 198,
    "Protoss Flag Beacon": 199,
    "Power Generator": 200,
    "Overmind Cocoon": 201,
    "Dark Swarm": 202,
    "Floor Missile Trap": 203,
    "Floor Hatch (Unused)": 204,
    "Floor Hatch (UNUSED)": 204,
    "Left Upper Level Door": 205,
    "Right Upper Level Door": 206,
    "Left Pit Door": 207,
    "Right Pit Door": 208,
    "Floor Gun Trap": 209,
    "Left Wall Missile Trap": 210,
    "Left Wall Flame Trap": 211,
    "Right Wall Missile Trap": 212,
    "Right Wall Flame Trap": 213,
    "Start Location": 214,
    "Flag": 215,
    "Young Chrysalis": 216,
    "Psi Emitter": 217,
    "Data Disc": 218,
    "Khaydarin Crystal": 219,
    "Mineral Chunk (Type 1)": 220,
    "Mineral Chunk (Type 2)": 221,
    "Mineral Cluster Type 1": 220,
    "Mineral Cluster Type 2": 221,
    "Vespene Orb (Protoss Type 1)": 222,
    "Vespene Orb (Protoss Type 2)": 223,
    "Protoss Vespene Gas Orb Type 1": 222,
    "Protoss Vespene Gas Orb Type 2": 223,
    "Vespene Sac (Zerg Type 1)": 224,
    "Vespene Sac (Zerg Type 2)": 225,
    "Zerg Vespene Gas Sac Type 1": 224,
    "Zerg Vespene Gas Sac Type 2": 225,
    "Vespene Tank (Terran Type 1)": 226,
    "Vespene Tank (Terran Type 2)": 227,
    "Terran Vespene Gas Tank Type 1": 226,
    "Terran Vespene Gas Tank Type 2": 227,
    "None": 228,
    "Cancel Order": 229,
    "Cancel Place COP": 230,
    "Building Is Constructed (Terran&Protoss)": 231,
    "Rallyable Building Is Constructed (Terran&Protoss)": 232,
    "Building Is Mutating (Zerg)": 233,
    "Rallyable Building Is Mutating (Zerg)": 234,
    "Cancel Infestation": 235,
    "Cancel Mutation (Lair&Hive)": 236,
    "Cancel Nuclear Strike": 237,
    "Zerg Basic Mutation": 238,
    "Terran Basic Structure": 239,
    "Protoss Basic Structure": 240,
    "Zerg Advanced Mutation": 241,
    "Terran Advanced Structure": 242,
    "Protoss Advanced Structure": 243,
    "Units": 244,
    "Workers": 245,
    "Cloakable Units": 246,
    "Burrowable Units": 247,
    "Replay (Play Button)": 248,
    "Replay (Pause Button)": 249,
}

DefaultButtonSet: TypeAlias = Literal[
    "Terran Marine",
    "Terran Ghost",
    "Terran Vulture",
    "Terran Goliath",
    "Goliath Turret",
    "Terran Siege Tank (Tank Mode)",
    "Siege Tank Turret (Tank Mode)",
    "Terran SCV",
    "Terran Wraith",
    "Terran Science Vessel",
    "Gui Montag",
    "Terran Dropship",
    "Terran Battlecruiser",
    "Vulture Spider Mine",
    "Nuclear Missile",
    "Terran Civilian",
    "Sarah Kerrigan",
    "Alan Schezar",
    "Alan Schezar Turret",
    "Jim Raynor (Vulture)",
    "Jim Raynor (Marine)",
    "Tom Kazansky",
    "Magellan",
    "Edmund Duke (Tank Mode)",
    "Edmund Duke Turret (Tank Mode)",
    "Edmund Duke (Siege Mode)",
    "Edmund Duke Turret (Siege Mode)",
    "Arcturus Mengsk",
    "Hyperion",
    "Norad II (Battlecruiser)",
    "Terran Siege Tank (Siege Mode)",
    "Siege Tank Turret (Siege Mode)",
    "Terran Firebat",
    "Scanner Sweep",
    "Terran Medic",
    "Zerg Larva",
    "Zerg Egg",
    "Zerg Zergling",
    "Zerg Hydralisk",
    "Zerg Ultralisk",
    "Zerg Broodling",
    "Zerg Drone",
    "Zerg Overlord",
    "Zerg Mutalisk",
    "Zerg Guardian",
    "Zerg Queen",
    "Zerg Defiler",
    "Zerg Scourge",
    "Torrasque",
    "Matriarch",
    "Infested Terran",
    "Infested Kerrigan",
    "Unclean One",
    "Hunter Killer",
    "Devouring One",
    "Kukulza (Mutalisk)",
    "Kukulza (Guardian)",
    "Yggdrasill",
    "Terran Valkyrie",
    "Cocoon",
    "Protoss Corsair",
    "Protoss Dark Templar",
    "Zerg Devourer",
    "Protoss Dark Archon",
    "Protoss Probe",
    "Protoss Zealot",
    "Protoss Dragoon",
    "Protoss High Templar",
    "Protoss Archon",
    "Protoss Shuttle",
    "Protoss Scout",
    "Protoss Arbiter",
    "Protoss Carrier",
    "Protoss Interceptor",
    "Dark Templar (Hero)",
    "Zeratul",
    "Tassadar/Zeratul",
    "Fenix (Zealot)",
    "Fenix (Dragoon)",
    "Tassadar",
    "Mojo" "Warbringer",
    "Gantrithor",
    "Protoss Reaver",
    "Protoss Observer",
    "Protoss Scarab",
    "Danimoth",
    "Aldaris",
    "Artanis",
    "Rhynadon",
    "Bengalaas",
    "Cargo Ship",
    "Mercenary Gunship",
    "Scantid",
    "Kakaru",
    "Ragnasaur",
    "Ursadon",
    "Lurker Egg",
    "Zerg Lurker Egg",
    "Raszagal",
    "Samir Duran",
    "Alexei Stukov",
    "Map Revealer",
    "Gerard DuGalle",
    "Zerg Lurker",
    "Infested Duran",
    "Disruption Web",
    "Terran Command Center",
    "Terran Comsat Station",
    "Terran Nuclear Silo",
    "Terran Supply Depot",
    "Terran Refinery",
    "Terran Barracks",
    "Terran Academy",
    "Terran Factory",
    "Terran Starport",
    "Terran Control Tower",
    "Terran Science Facility",
    "Terran Covert Ops",
    "Terran Physics Lab",
    "Starbase",
    "Terran Machine Shop",
    "Repair Bay",
    "Terran Engineering Bay",
    "Terran Armory",
    "Terran Missile Turret",
    "Terran Bunker",
    "Norad II (Crashed)",
    "Ion Cannon",
    "Uraj Crystal",
    "Khalis Crystal",
    "Infested Command Center",
    "Zerg Hatchery",
    "Zerg Lair",
    "Zerg Hive",
    "Zerg Nydus Canal",
    "Zerg Hydralisk Den",
    "Zerg Defiler Mound",
    "Zerg Greater Spire",
    "Zerg Queen's Nest",
    "Zerg Evolution Chamber",
    "Zerg Ultralisk Cavern",
    "Zerg Spire",
    "Zerg Spawning Pool",
    "Zerg Creep Colony",
    "Zerg Spore Colony",
    "Unused Zerg Building1",
    "Zerg Sunken Colony",
    "Zerg Overmind (With Shell)",
    "Zerg Overmind",
    "Zerg Extractor",
    "Mature Chrysalis",
    "Zerg Cerebrate",
    "Zerg Cerebrate Daggoth",
    "Unused Zerg Building2",
    "Protoss Nexus",
    "Protoss Robotics Facility",
    "Protoss Pylon",
    "Protoss Assimilator",
    "Unused Protoss Building1",
    "Protoss Observatory",
    "Protoss Gateway",
    "Unused Protoss Building2",
    "Protoss Photon Cannon",
    "Protoss Citadel of Adun",
    "Protoss Cybernetics Core",
    "Protoss Templar Archives",
    "Protoss Forge",
    "Protoss Stargate",
    "Stasis Cell/Prison",
    "Protoss Fleet Beacon",
    "Protoss Arbiter Tribunal",
    "Protoss Robotics Support Bay",
    "Protoss Shield Battery",
    "Khaydarin Crystal Formation",
    "Protoss Temple",
    "Xel'Naga Temple",
    "Mineral Field (Type 1)",
    "Mineral Field (Type 2)",
    "Mineral Field (Type 3)",
    "Cave",
    "Cave-in",
    "Cantina",
    "Mining Platform",
    "Independent Command Center",
    "Independent Starport",
    "Jump Gate",
    "Ruins",
    "Khaydarin Crystal Formation (Unused)",
    "Vespene Geyser",
    "Warp Gate",
    "Psi Disrupter",
    "Zerg Marker",
    "Terran Marker",
    "Protoss Marker",
    "Zerg Beacon",
    "Terran Beacon",
    "Protoss Beacon",
    "Zerg Flag Beacon",
    "Terran Flag Beacon",
    "Protoss Flag Beacon",
    "Power Generator",
    "Overmind Cocoon",
    "Dark Swarm",
    "Floor Missile Trap",
    "Floor Hatch (Unused)",
    "Left Upper Level Door",
    "Right Upper Level Door",
    "Left Pit Door",
    "Right Pit Door",
    "Floor Gun Trap",
    "Left Wall Missile Trap",
    "Left Wall Flame Trap",
    "Right Wall Missile Trap",
    "Right Wall Flame Trap",
    "Start Location",
    "Flag",
    "Young Chrysalis",
    "Psi Emitter",
    "Data Disc",
    "Khaydarin Crystal",
    "Mineral Chunk (Type 1)",
    "Mineral Chunk (Type 2)",
    "Vespene Orb (Protoss Type 1)",
    "Vespene Orb (Protoss Type 2)",
    "Vespene Sac (Zerg Type 1)",
    "Vespene Sac (Zerg Type 2)",
    "Vespene Tank (Terran Type 1)",
    "Vespene Tank (Terran Type 2)",
    "None",
    "Cancel Order",
    "Cancel Place COP",
    "Building Is Constructed (Terran&Protoss)",
    "Rallyable Building Is Constructed (Terran&Protoss)",
    "Building Is Mutating (Zerg)",
    "Rallyable Building Is Mutating (Zerg)",
    "Cancel Infestation",
    "Cancel Mutation (Lair&Hive)",
    "Cancel Nuclear Strike",
    "Zerg Basic Mutation",
    "Terran Basic Structure",
    "Protoss Basic Structure",
    "Zerg Advanced Mutation",
    "Terran Advanced Structure",
    "Protoss Advanced Structure",
    "Units",
    "Workers",
    "Cloakable Units",
    "Burrowable Units",
    "Replay (Play Button)",
    "Replay (Pause Button)",
]



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from typing import Literal

from ...utils import (
    EUDCreateBlock,
    EUDGetLastBlockOfName,
    EUDPopBlock,
)
from ..allocator import ConstExpr, Forward


def PushTriggerScope() -> Literal[True]:  # noqa: N802
    EUDCreateBlock("triggerscope", {"nexttrigger_list": []})
    return True  # Allow `if PushTriggerScope()` syntax for indent


def SetNextTrigger(trg: ConstExpr) -> None:  # noqa: N802
    """For optimization purpose, one may call this function directly"""
    nt_list = EUDGetLastBlockOfName("triggerscope")[1]["nexttrigger_list"]
    for fw in nt_list:
        fw << trg
    nt_list.clear()


def NextTrigger() -> Forward:  # noqa: N802
    fw = Forward()
    nt_list = EUDGetLastBlockOfName("triggerscope")[1]["nexttrigger_list"]
    nt_list.append(fw)
    return fw


def _register_trigger(trg: ConstExpr) -> None:
    SetNextTrigger(trg)


def PopTriggerScope() -> None:  # noqa: N802
    nt_list = EUDPopBlock("triggerscope")[1]["nexttrigger_list"]
    for fw in nt_list:
        fw << 0



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

# This code uses simple LCG algorithm.

from ... import core as c
from ... import ctrlstru as cs
from ... import utils as ut
from ...core.eudfunc.eudf import _EUDPredefineReturn
from ...core.variable.vbase import VariableBase
from ...memio.rwcommon import lv

_seed = c.EUDVariable()


def f_getseed() -> c.EUDVariable:
    t = c.EUDVariable()
    t << _seed
    return t


def f_srand(seed) -> None:
    _seed << seed


@c.EUDFunc
def f_randomize():
    global _seed

    # Store switch 1
    end = c.Forward()
    n = lv
    c.RawTrigger(
        actions=[
            _seed.SetNumber(0),
            _seed.QueueAddTo(_seed),
            c.SetMemoryX(end + 328 + 24, c.SetTo, 5 << 24, 0xFF << 24),
            n.SetNumber(32),
        ]
    )
    c.RawTrigger(
        conditions=c.Switch("Switch 1", c.Set),
        actions=c.SetMemoryX(end + 328 + 24, c.SetTo, 4 << 24, 0xFF << 24),
    )

    if cs.EUDWhile()(n >= 1):
        # _seed += _seed
        c.VProc(
            _seed,
            [
                c.SetSwitch("Switch 1", c.Random),
                n.SubtractNumber(1),
            ],
        )
        c.RawTrigger(
            conditions=c.Switch("Switch 1", c.Set), actions=_seed.AddNumber(1)
        )
    cs.EUDEndWhile()

    end << c.RawTrigger(actions=c.SetSwitch("Switch 1", c.Set))


@_EUDPredefineReturn(2, 3)
@c.EUDFunc
def f_rand():
    ret = f_rand._frets[0]
    c.f_mul(_seed, 1103515245, ret=[_seed])
    c.VProc(_seed, [_seed.AddNumber(12345), _seed.QueueAssignTo(ret)])
    ret >>= 16
    # Only HIWORD is returned
    # return ret


@c.EUDFunc
def f_dwrand():
    dseed = c.f_mul(_seed, 1103515245)
    dseed += 12345
    c.f_mul(dseed, 1103515245, ret=[_seed])
    c.RawTrigger(
        actions=[
            _seed.AddNumber(12345),
            # HIWORD
            dseed.SetNumberX(0, 0xFFFF),
        ],
    )

    # LOWORD
    for i in ut._rand_lst(range(16, 32)):
        c.RawTrigger(
            conditions=_seed.AtLeastX(1, 2**i),
            actions=dseed.AddNumber(2 ** (i - 16)),
        )

    return dseed



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from ...localize import _
from ...utils import EPError, ExprProxy
from .selftype import selftype
from .vararray import _InternalVArray


class _EUDStructMetaclass(type):
    def __init__(cls, name, bases, dct) -> None:
        # For field declaration, modify selftype to cls
        fielddict = {}
        fieldcount = 0
        for basetype in reversed(cls.__mro__):
            if not hasattr(basetype, "_fields_"):
                continue
            for i, nametype in enumerate(basetype._fields_):
                if isinstance(nametype, str):  # "fieldname"
                    fieldname = nametype
                    fieldtype = None
                else:  # ("fieldname", fieldtype)
                    fieldname = nametype[0]
                    fieldtype = nametype[1]
                if fieldtype is selftype:
                    fieldtype = cls
                    if not isinstance(basetype._fields_, list):
                        basetype._fields_ = list(basetype._fields_)
                    basetype._fields_[i] = (fieldname, cls)

                if fieldname in fielddict:
                    raise EPError(_("Duplicated field name: {}").format(fieldname))
                fielddict[fieldname] = (fieldcount, fieldtype)
                fieldcount += 1

        cls._fielddict = fielddict
        super().__init__(name, bases, dct)

    def __mul__(cls, times: int):
        basetype = cls
        return _EUDStructArray(times, basetype)


class EUDStructArray(ExprProxy, metaclass=_EUDStructMetaclass):
    dont_flatten = True

    def __init__(self, initvar=None, *, _from=None, _times, _basetype):
        if _from is None:
            if initvar is None:
                initvals = [0] * _times
                super().__init__(_InternalVArray(_times, _basetype)(initvals))
            else:
                super().__init__(_InternalVArray(_times, _basetype)(initvar))
        else:
            super().__init__(_InternalVArray(_times, _basetype).cast(_from))

        self._initialized = True
        self._times = _times

    def copy(self):
        """Create a shallow copy"""
        arraytype = type(self)
        inst = arraytype()
        self.copyto(inst)
        return inst

    def copyto(self, inst):
        """Copy struct to other instance"""
        for i in range(self._times):
            inst[i] = self[i]

    def __getitem__(self, index):
        if isinstance(index, int) and not 0 <= index < self._times:
            e = _(
                "index out of bounds: the length of EUDVArray is {}"
                " but the index is {}"
            )
            raise EPError(e.format(self._times, index))

        return self.getValue()[index]

    def __setitem__(self, index, newval):
        if isinstance(index, int) and not 0 <= index < self._times:
            e = _(
                "index out of bounds: the length of EUDVArray is {}"
                " but the index is {}"
            )
            raise EPError(e.format(self._times, index))
        self.getValue()[index] = newval

    def __getattr__(self, name):
        return super().__getattr__(name)

    def __setattr__(self, name, value):
        self.__dict__[name] = value


class _EUDStructArray:
    def __init__(self, times: int, basetype) -> None:
        self._times: int = times
        self._basetype = basetype

    def __call__(self, initvar=None, *, _from=None) -> EUDStructArray:
        return EUDStructArray(
            initvar=initvar,
            _from=_from,
            _times=self._times,
            _basetype=self._basetype,
        )

    def cast(self, _from):
        return self(_from=_from)

    def __mul__(self, times: int):
        return _EUDStructArray(times, self)



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from ..utils import EPD
from .allocator import Forward
from .rawtrigger import Add, EncodePlayer, Exactly, Memory, SetMemory, SetTo
from .variable import EUDXVariable

_curpl_var = EUDXVariable(EPD(0x6509B0), SetTo, 0)
_curpl_checkcond = Forward()


def cpcache_match_cond():
    if not _curpl_checkcond.IsSet():
        _curpl_checkcond << Memory(0x6509B0, Exactly, 0)
    return _curpl_checkcond


def SetCurrentPlayer(p):  # noqa: N802
    p = EncodePlayer(p)
    return [
        _curpl_var.SetNumber(p),
        SetMemory(_curpl_checkcond + 8, SetTo, p),
        SetMemory(0x6509B0, SetTo, p),
    ]


def AddCurrentPlayer(p):  # noqa: N802
    p = EncodePlayer(p)
    return [
        _curpl_var.AddNumber(p),
        SetMemory(_curpl_checkcond + 8, Add, p),
        SetMemory(0x6509B0, Add, p),
    ]


def GetCPCache():  # noqa: N802
    return _curpl_var



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.


class Payload:
    def __init__(
        self, data: bytes, prttable: list[int], orttable: list[int]
    ) -> None:
        self.data = data
        self.prttable = prttable
        self.orttable = orttable



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from itertools import chain, combinations
from typing import Literal

from .. import core as c
from .. import trigger as tg
from .. import utils as ut
from ..localize import _
from .basicstru import EUDJump, EUDJumpIf, EUDJumpIfNot
from .cshelper import CtrlStruOpener
from .jumptable import JumpTriggerForward

cpcache = c.curpl.GetCPCache()


def _is_switch_blockid(idf) -> bool:
    return idf == "swblock"


def EPDSwitch(epd, mask=0xFFFFFFFF) -> Literal[True]:  # noqa: N802
    epd = c.EncodePlayer(epd)
    epd = ut.unProxy(epd)
    block = {
        "targetepd": epd,
        "bitmask": mask,
        "casebrlist": {},
        "defaultbr": c.Forward(),
        "swend": c.Forward(),
        "_actions": [],
    }

    c.PushTriggerScope()
    ut.EUDCreateBlock("swblock", block)

    return True


def EUDSwitch(var, mask=0xFFFFFFFF) -> Literal[True]:  # noqa: N802
    var = ut.unProxy(var)
    block = {
        "targetvar": var,
        "bitmask": mask,
        "casebrlist": {},
        "defaultbr": c.Forward(),
        "swend": c.Forward(),
        "_actions": [],
    }

    c.PushTriggerScope()
    ut.EUDCreateBlock("swblock", block)

    return True


def EUDSwitchCase() -> CtrlStruOpener:  # noqa: N802
    def _footer(*numbers: int | c.ConstExpr) -> Literal[True]:
        for number in numbers:
            ut.ep_assert(
                isinstance(number, int) or isinstance(number, c.ConstExpr),
                _("Invalid selector start for EUDSwitch"),
            )

        lb = ut.EUDGetLastBlock()
        ut.ep_assert(lb[0] == "swblock", _("Block start/end mismatch"))
        block = lb[1]

        for number in numbers:
            case = number & block["bitmask"]
            ut.ep_assert(
                block["bitmask"] == 0xFFFFFFFF or case == number,
                _("case out of bitmask: {0} & 0x{1:X} != {0}").format(
                    number, block["bitmask"]
                ),
            )
            ut.ep_assert(case not in block["casebrlist"], _("Duplicate cases"))
            block["casebrlist"][case] = c.NextTrigger()

        return True

    return CtrlStruOpener(_footer)


def EUDSwitchDefault() -> CtrlStruOpener:  # noqa: N802
    def _footer() -> Literal[True]:
        lb = ut.EUDGetLastBlock()
        ut.ep_assert(lb[0] == "swblock", _("Block start/end mismatch"))
        block = lb[1]

        ut.ep_assert(not block["defaultbr"].IsSet(), _("Duplicate default"))
        block["defaultbr"] << c.NextTrigger()

        return True

    return CtrlStruOpener(_footer)


def eudswitch_break() -> None:
    block = ut.EUDGetLastBlockOfName("swblock")[1]
    try:
        EUDJump(block["swend"])
    except ut.EPError:
        ut.ep_warn(_("unreachable break"))


def eudswitch_break_if(conditions) -> None:
    block = ut.EUDGetLastBlockOfName("swblock")[1]
    EUDJumpIf(conditions, block["swend"])


def eudswitch_break_ifnot(conditions) -> None:
    block = ut.EUDGetLastBlockOfName("swblock")[1]
    EUDJumpIfNot(conditions, block["swend"])


def EUDEndSwitch() -> None:  # noqa: N802
    lb = ut.EUDPopBlock("swblock")
    block = lb[1]
    swend = block["swend"]
    c.SetNextTrigger(swend)  # Exit switch block
    c.PopTriggerScope()

    bitmask: int = block["bitmask"]
    casebrlist: dict[int, c.Forward] = block["casebrlist"]
    defbranch: c.Forward = block["defaultbr"]
    casekeylist: list[int] = sorted(block["casebrlist"].keys())

    # If default block is not specified, skip it.
    if not defbranch.IsSet():
        defbranch << swend

    # If there is only the default destination, jump there directly.
    if not casekeylist:
        c.SetNextTrigger(defbranch)
        swend << c.NextTrigger()
        return

    try:
        epd = ut.EPD(block["targetvar"].getValueAddr())
    except KeyError:
        epd = block["targetepd"]
    except AttributeError:
        # constant value switch
        if block["targetvar"] in casekeylist:
            c.SetNextTrigger(casebrlist[block["targetvar"]])
        else:
            c.SetNextTrigger(defbranch)
        swend << c.NextTrigger()
        return

    if (
        len(casekeylist) == 2
        and casebrlist[casekeylist[0]] == casebrlist[casekeylist[1]]
    ):
        keyand = casekeylist[0] & casekeylist[1]
        keyxor = (casekeylist[0] | casekeylist[1]) - keyand
        if keyxor.bit_count() == 1:
            tg.EUDBranch(
                c.MemoryXEPD(epd, c.Exactly, keyand, ~keyxor & bitmask),
                casebrlist[casekeylist[0]],
                defbranch,
            )
            swend << c.NextTrigger()
            return

    if len(casekeylist) <= 3 and not c.IsEUDVariable(epd):
        # use simple comparisons
        branch, nextbranch = c.Forward(), c.Forward()
        restore = c.SetMemory(branch + 4, c.SetTo, nextbranch)
        c.RawTrigger(actions=[restore, *block["_actions"]])
        for case in casekeylist:
            branch << c.RawTrigger(
                conditions=c.MemoryXEPD(epd, c.Exactly, case, bitmask),
                actions=[
                    c.SetNextPtr(branch, casebrlist[case]),
                    c.SetMemory(restore + 16, c.SetTo, ut.EPD(branch) + 1),
                    c.SetMemory(restore + 20, c.SetTo, nextbranch),
                ],
            )
            nextbranch << c.NextTrigger()
            branch, nextbranch = c.Forward(), c.Forward()
        c.SetNextTrigger(defbranch)
        swend << c.NextTrigger()
        return

    keyand, keyor = casekeylist[0], 0
    for key in casekeylist:
        keyand &= key
        keyor |= key
    keyxor = keyor - keyand
    keybits = list(ut.bits(keyxor))

    density = len(casebrlist) / (2 ** len(keybits))
    if density >= 0.4:
        # use jump table
        if (~keyor | keyand) & bitmask != 0:
            check_invariant = c.MemoryXEPD(
                epd, c.Exactly, keyand, (~keyor | keyand) & bitmask
            )
            EUDJumpIfNot(check_invariant, defbranch)

        def powerset(iterable) -> chain[tuple]:
            "powerset([1,2,3]) --> () (1,) (2,) (3,) (1,2) (1,3) (2,3) (1,2,3)"
            s = list(iterable)
            return chain.from_iterable(combinations(s, r) for r in range(len(s) + 1))

        keylist = sorted(map(sum, powerset(keybits)))
        jump_table = JumpTriggerForward(
            [casebrlist.get(keyand + key, defbranch) for key in keylist]
        )
        jumper = c.Forward()

        cmpplayer = epd
        if c.IsEUDVariable(epd):
            cmpplayer = 13  # CurrentPlayer
            c.VProc(
                epd,
                [
                    cpcache.SetDest(ut.EPD(0x6509B0)),
                    *epd.QueueAssignTo(ut.EPD(0x6509B0)),
                    c.SetNextPtr(jumper, jump_table),
                    *block["_actions"],
                ],
            )
        else:
            restore = c.SetNextPtr(jumper, jump_table)
            lastbit = c.RawTrigger(actions=[restore, *block["_actions"]])

        for i, bit in enumerate(keybits):
            lastbit = c.RawTrigger(
                conditions=c.MemoryXEPD(cmpplayer, c.AtLeast, 1, bit),
                actions=c.SetMemory(jumper + 4, c.Add, 20 * (1 << i)),
            )

        if c.IsEUDVariable(epd):
            jumper << cpcache.GetVTable()
            c.SetNextTrigger(cpcache.GetVTable())
        else:
            jumper << lastbit
    elif c.IsEUDVariable(epd):
        # use binary search with CP trick
        cmpplayer = 13  # CurrentPlayer
        c.VProc(
            epd,
            [
                cpcache.SetDest(ut.EPD(0x6509B0)),
                *epd.QueueAssignTo(ut.EPD(0x6509B0)),
                c.SetNextPtr(cpcache.GetVTable(), defbranch),
                *block["_actions"],
            ],
        )
        reset: list[tuple[c.Forward, c.Forward]] = []

        def _reset():
            nonlocal reset
            return [c.SetNextPtr(trg, nptr) for trg, nptr in reset]

        def _key_selector(keys: list[int]) -> c.Forward:
            nonlocal reset
            ret = c.NextTrigger()
            if len(keys) == 1:  # Only one keys on the list
                if reset:
                    c.RawTrigger(actions=_reset())
                c.RawTrigger(
                    nextptr=cpcache.GetVTable(),
                    conditions=c.MemoryXEPD(cmpplayer, c.Exactly, keys[0], bitmask),
                    actions=c.SetNextPtr(cpcache.GetVTable(), casebrlist[keys[0]]),
                )

            elif len(keys) == 2:
                br1, jump1, br2 = c.Forward(), c.Forward(), c.Forward()
                br1 << c.RawTrigger(
                    nextptr=br2,
                    conditions=c.MemoryXEPD(cmpplayer, c.Exactly, keys[0], bitmask),
                    actions=[c.SetNextPtr(br1, jump1), *_reset()],
                )
                jump1 << c.RawTrigger(
                    nextptr=cpcache.GetVTable(),
                    actions=[
                        c.SetNextPtr(cpcache.GetVTable(), casebrlist[keys[0]]),
                        c.SetNextPtr(br1, br2),
                    ],
                )
                br2 << _key_selector(keys[1:])

            elif len(keys) >= 3:
                branch, br1, br2 = c.Forward(), c.Forward(), c.Forward()
                midpos = len(keys) // 2
                branch << c.RawTrigger(
                    nextptr=br1,
                    conditions=c.MemoryXEPD(
                        cmpplayer, c.AtLeast, keys[midpos], bitmask
                    ),
                    actions=[
                        c.SetNextPtr(branch, br2),
                        *_reset(),
                    ],
                )
                br1 << _key_selector(keys[:midpos])
                reset.clear()
                reset.append((branch, br1))
                br2 << _key_selector(keys[midpos:])

            else:  # len(keys) == 0
                return defbranch

            return ret

        _key_selector(casekeylist)
    else:  # use binary search
        reset = []

        def _reset():
            nonlocal reset
            return [c.SetNextPtr(trg, nptr) for trg, nptr in reset]

        def _key_selector(keys: list[int]) -> c.Forward:
            nonlocal reset
            ret = c.NextTrigger()
            if len(keys) == 1:  # Only one keys on the list
                branch, jump, goto_defbranch = (
                    c.Forward(),
                    c.Forward(),
                    c.Forward(),
                )
                branch << c.RawTrigger(
                    nextptr=goto_defbranch,
                    conditions=c.MemoryXEPD(epd, c.Exactly, keys[0], bitmask),
                    actions=[c.SetNextPtr(branch, jump), *_reset()],
                )
                jump << c.RawTrigger(
                    nextptr=casebrlist[keys[0]],
                    actions=c.SetNextPtr(branch, goto_defbranch),
                )
                if reset:
                    goto_defbranch << c.RawTrigger(
                        nextptr=defbranch,
                        actions=_reset(),
                    )
                else:
                    goto_defbranch << defbranch

            elif len(keys) == 2:
                br1, jump1, br2 = c.Forward(), c.Forward(), c.Forward()
                br1 << c.RawTrigger(
                    nextptr=br2,
                    conditions=c.MemoryXEPD(epd, c.Exactly, keys[0], bitmask),
                    actions=[c.SetNextPtr(br1, jump1), *_reset()],
                )
                jump1 << c.RawTrigger(
                    nextptr=casebrlist[keys[0]],
                    actions=c.SetNextPtr(br1, br2),
                )
                br2 << _key_selector(keys[1:])

            elif len(keys) >= 3:
                branch, br1, br2 = c.Forward(), c.Forward(), c.Forward()
                midpos = len(keys) // 2
                branch << c.RawTrigger(
                    nextptr=br1,
                    conditions=c.MemoryXEPD(epd, c.AtLeast, keys[midpos], bitmask),
                    actions=[
                        c.SetNextPtr(branch, br2),
                        *_reset(),
                    ],
                )
                br1 << _key_selector(keys[:midpos])
                reset.clear()
                reset.append((branch, br1))
                br2 << _key_selector(keys[midpos:])

            else:  # len(keys) == 0
                return defbranch

            return ret

        _key_selector(casekeylist)

    swend << c.NextTrigger()



# Copyright 2023 by Armoha.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

# ruff: noqa: I001
# fmt: off
from .csprite import CSprite
from .cunit import CUnit, EPDCUnitMap
from .flingy import Flingy
from .image import Image
from .player import (
    TrgPlayer,
    P1, P2, P3, P4, P5, P6, P7, P8, P9, P10, P11, P12,
    Player1, Player2, Player3, Player4, Player5, Player6,
    Player7, Player8, Player9, Player10, Player11, Player12,
    Force1, Force2, Force3, Force4, AllPlayers, CurrentPlayer,
    Allies, Foes, NeutralPlayers, NonAlliedVictoryPlayers,
)
from .sprite import Sprite
from .tech import Tech
from .unit import TrgUnit
from .unitorder import UnitOrder
from .upgrade import Upgrade
from .weapon import Weapon

__all__ = [
    "CSprite",
    "CUnit",
    "EPDCUnitMap",
    "Flingy",
    "Image",
    "TrgPlayer",
    "P1", "P2", "P3", "P4", "P5", "P6", "P7", "P8", "P9", "P10", "P11", "P12",
    "Player1", "Player2", "Player3", "Player4", "Player5", "Player6",
    "Player7", "Player8", "Player9", "Player10", "Player11", "Player12",
    "Force1", "Force2", "Force3", "Force4", "AllPlayers", "CurrentPlayer",
    "Allies", "Foes", "NeutralPlayers", "NonAlliedVictoryPlayers",
    "Sprite",
    "Tech",
    "TrgUnit",
    "UnitOrder",
    "Upgrade",
    "Weapon",
]
# fmt: on



from .. import core as c
from .. import utils as ut
from ..core.eudfunc import EUDFullFunc
from ..eudlib.utilf.userpl import (
    DisplayTextAll,
    f_getuserplayerid,
)
from ..memio import f_getcurpl, f_setcurpl, f_setcurpl2cpcache
from .cpprint import FixedText
from .strbuffer import GetGlobalStringBuffer


def f_printAll(format_string, *args):  # noqa: N802
    if not args:
        return f_setcurpl2cpcache([], DisplayTextAll(format_string))
    oldcp = f_getcurpl()
    f_setcurpl(f_getuserplayerid())
    gsb = GetGlobalStringBuffer()
    gsb.printf(format_string, *args)
    f_setcurpl(oldcp)


_display_text_all = DisplayTextAll(0)


@EUDFullFunc(
    [
        (ut.EPD(0x640B58), c.Add, 0),
        (ut.EPD(_display_text_all[1]) + 1, c.SetTo, 0),
    ],
    [None, c.TrgString],
)
def DisplayTextAllAt(line, text):  # noqa: N802
    with FixedText(_display_text_all):
        c.VProc([line, text], [])
        c.RawTrigger(
            conditions=c.Memory(0x640B58, c.AtLeast, 11),
            actions=c.SetMemory(0x640B58, c.Subtract, 11),
        )
    f_setcurpl2cpcache()


def f_printAllAt(line, format_string, *args):  # noqa: N802
    if not args:
        DisplayTextAllAt(line, format_string)
    oldcp = f_getcurpl()
    f_setcurpl(f_getuserplayerid())
    gsb = GetGlobalStringBuffer()
    gsb.printfAt(line, format_string, *args)
    f_setcurpl(oldcp)
